from __future__ import annotations
import ast
import sys
from skylos.rules.danger.taint import TaintVisitor


def _qualified_name_from_call(node):
    func = node.func
    parts = []
    while isinstance(func, ast.Attribute):
        parts.append(func.attr)
        func = func.value
    if isinstance(func, ast.Name):
        parts.append(func.id)
        parts.reverse()
        return ".".join(parts)
    if isinstance(func, ast.Name):
        return func.id
    return None


def _has_safe_base_url(node):

    if not node.values:
        return True
    
    first = node.values[0]

    if isinstance(first, ast.Constant) and isinstance(first.value, str):
        val = first.value

        if "://" in val:
            parts = val.split("://", 1)
            if len(parts) > 1 and "/" in parts[1]:
                return True

    if isinstance(first, ast.FormattedValue):
        if isinstance(first.value, ast.Name) and first.value.id.isupper():
            return True
            
    return False


def _is_interpolated_string(node):
    if isinstance(node, ast.JoinedStr):
        if _has_safe_base_url(node):
            return False
        return True

    if isinstance(node, ast.BinOp) and isinstance(node.op, (ast.Add, ast.Mod)):
        return True

    if (
        isinstance(node, ast.Call)
        and isinstance(node.func, ast.Attribute)
        and node.func.attr == "format"
    ):
        return True
    return False


class _SSRFFlowChecker(TaintVisitor):
    HTTP_METHODS = {"get", "post", "put", "delete", "head", "options", "request"}

    def visit_Call(self, node):
        qn = _qualified_name_from_call(node)

        if qn and "." in qn:
            _, func = qn.split(".", 1)
            if func in self.HTTP_METHODS and node.args:
                url_arg = node.args[0]
                if _is_interpolated_string(url_arg) or self.is_tainted(url_arg):
                    self.findings.append(
                        {
                            "rule_id": "SKY-D216",
                            "severity": "CRITICAL",
                            "message": "Possible SSRF: tainted URL passed to HTTP client.",
                            "file": str(self.file_path),
                            "line": node.lineno,
                            "col": node.col_offset,
                        }
                    )

        if qn and qn.endswith(".urlopen") and node.args:
            url_arg = node.args[0]
            if _is_interpolated_string(url_arg) or self.is_tainted(url_arg):
                self.findings.append(
                    {
                        "rule_id": "SKY-D216",
                        "severity": "CRITICAL",
                        "message": "Possible SSRF: tainted URL passed to HTTP client.",
                        "file": str(self.file_path),
                        "line": node.lineno,
                        "col": node.col_offset,
                    }
                )

        self.generic_visit(node)


def scan(tree, file_path, findings):
    try:
        checker = _SSRFFlowChecker(file_path, findings)
        checker.visit(tree)
    except Exception as e:
        print(f"SSRF flow analysis failed for {file_path}: {e}", file=sys.stderr)



from __future__ import annotations
import ast
from pathlib import Path
from skylos.rules.base import SkylosRule


class PerformanceRule(SkylosRule):
    rule_id = "SKY-P401"
    name = "Performance Checks"

    def __init__(self, ignore_list=None):
        self.ignore_list = ignore_list or []

    def _is_pandas_read(self, node):
        if isinstance(node.func, ast.Attribute) and node.func.attr == "read_csv":
            return True
        return False

    def visit_node(self, node, context):
        findings = []

        if isinstance(node, ast.Call):
            if isinstance(node.func, ast.Attribute) and node.func.attr in (
                "read",
                "readlines",
            ):
                if "SKY-P401" not in self.ignore_list:
                    findings.append(
                        {
                            "rule_id": "SKY-P401",
                            "kind": "performance",
                            "severity": "MEDIUM",
                            "type": "function",
                            "name": node.func.attr,
                            "simple_name": node.func.attr,
                            "value": "memory_load",
                            "threshold": 0,
                            "message": f"Potential Memory Risk: '{node.func.attr}()' loads entire file into RAM. Consider iterating line-by-line.",
                            "file": context.get("filename"),
                            "basename": Path(context.get("filename", "")).name,
                            "line": node.lineno,
                            "col": node.col_offset,
                        }
                    )

            if self._is_pandas_read(node):
                if "SKY-P402" not in self.ignore_list:
                    has_chunk = False
                    for kw in node.keywords:
                        if kw.arg == "chunksize":
                            has_chunk = True
                            break

                    if not has_chunk:
                        findings.append(
                            {
                                "rule_id": "SKY-P402",
                                "kind": "performance",
                                "severity": "LOW",
                                "type": "function",
                                "name": "read_csv",
                                "simple_name": "read_csv",
                                "value": "no_chunk",
                                "threshold": 0,
                                "message": "Pandas Memory Risk: read_csv used without 'chunksize'. Large files may crash RAM.",
                                "file": context.get("filename"),
                                "basename": Path(context.get("filename", "")).name,
                                "line": node.lineno,
                                "col": node.col_offset,
                            }
                        )

        if isinstance(node, ast.For):
            if "SKY-P403" not in self.ignore_list:
                for child in node.body:
                    if isinstance(child, ast.For):
                        findings.append(
                            {
                                "rule_id": "SKY-P403",
                                "kind": "performance",
                                "severity": "LOW",
                                "type": "loop",
                                "name": "nested_loop",
                                "simple_name": "for",
                                "value": "O(N^2)",
                                "threshold": 0,
                                "message": "Performance Warning: Nested loop detected (O(N^2) complexity).",
                                "file": context.get("filename"),
                                "basename": Path(context.get("filename", "")).name,
                                "line": child.lineno,
                                "col": child.col_offset,
                            }
                        )

        return findings if findings else None



from .core import TypeScriptCore
from .danger import scan_danger
from .quality import scan_quality


class DummyVisitor:
    def __init__(self):
        self.is_test_file = False
        self.test_decorated_lines = set()
        self.dataclass_fields = set()
        self.pydantic_models = set()
        self.class_defs = {}
        self.first_read_lineno = {}
        self.framework_decorated_lines = set()


def scan_typescript_file(file_path, config=None):
    if config is None:
        config = {}

    try:
        with open(file_path, "rb") as f:
            source = f.read()
    except Exception:
        return ([], [], set(), set(), DummyVisitor(), DummyVisitor(), [], [], [])

    complexity_limit = config.get("complexity", 10)

    lang_overrides = config.get("languages", {}).get("typescript", {})
    complexity_limit = lang_overrides.get("complexity", complexity_limit)

    core = TypeScriptCore(file_path, source)
    core.scan()

    d_findings = scan_danger(core.root_node, file_path)
    q_findings = scan_quality(
        core.root_node, source, file_path, threshold=complexity_limit
    )

    return (
        core.defs,
        core.refs,
        set(),
        set(),
        DummyVisitor(),
        DummyVisitor(),
        q_findings,
        d_findings,
        [],
    )



from pathlib import Path

DEFAULTS = {
    "complexity": 10,
    "nesting": 3,
    "max_args": 5,
    "max_lines": 50,
    "ignore": [],
}


def load_config(start_path):
    current = Path(start_path).resolve()
    if current.is_file():
        current = current.parent

    root_config = None

    while True:
        toml_path = current / "pyproject.toml"
        if toml_path.exists():
            root_config = toml_path
            break
        if current.parent == current:
            break
        current = current.parent

    if not root_config:
        return DEFAULTS.copy()

    try:
        import tomllib  # pragma: no skylos
    except ImportError:
        try:
            import tomli as tomllib
        except ImportError:
            return DEFAULTS.copy()

    try:
        with open(root_config, "rb") as f:
            data = tomllib.load(f)

        user_cfg = data.get("tool", {}).get("skylos", {})

        gate_cfg = data.get("tool", {}).get("skylos", {}).get("gate", {})
        user_cfg["gate"] = gate_cfg

        final_cfg = DEFAULTS.copy()
        final_cfg.update(user_cfg)
        return final_cfg

    except Exception:
        return DEFAULTS.copy()



import datetime
import uuid
import hashlib  # Unused import
from typing import List, Dict, Optional

from .utils import get_current_time


class BaseModel:
    """Base model class with common functionality."""

    def __init__(self, id=None):
        self.id = id or str(uuid.uuid4())
        self.created_at = get_current_time()

    def to_dict(self):
        """Convert model to dictionary - used method."""
        return {"id": self.id, "created_at": self.created_at}

    def to_json(self):
        """Convert model to JSON - unused method."""
        import json

        return json.dumps(self.to_dict())

    @classmethod
    def from_dict(cls, data):
        """Create instance from dictionary - used method."""
        instance = cls(id=data.get("id"))
        if "created_at" in data:
            instance.created_at = data["created_at"]
        return instance


class User(BaseModel):
    """User model - used class."""

    def __init__(self, username, email, id=None):
        super().__init__(id)
        self.username = username
        self.email = email
        self.active = True

    def to_dict(self):
        """Override to_dict to include user fields - used method."""
        data = super().to_dict()
        data.update(
            {"username": self.username, "email": self.email, "active": self.active}
        )
        return data

    def deactivate(self):
        """Deactivate user - used method."""
        self.active = False

    def send_welcome_email(self):
        """Send welcome email - unused method."""
        pass


class Product(BaseModel):
    """Product model - used class."""

    def __init__(self, name, price, id=None):
        super().__init__(id)
        self.name = name
        self.price = price

    def to_dict(self):
        """Override to_dict for product - used method."""
        data = super().to_dict()
        data.update({"name": self.name, "price": self.price})
        return data

    def apply_discount(self, percentage):
        """Apply discount to product - unused method."""
        self.price = self.price * (1 - percentage / 100)


class Order(BaseModel):
    """Order model - unused class."""

    def __init__(self, user_id, products, id=None):
        super().__init__(id)
        self.user_id = user_id
        self.products = products
        self.status = "pending"

    def to_dict(self):
        """Order to dict - unused method."""
        data = super().to_dict()
        data.update(
            {"user_id": self.user_id, "products": self.products, "status": self.status}
        )
        return data

    def complete(self):
        """Mark order as complete - unused method."""
        self.status = "completed"


class Cart:
    """Shopping cart - unused class."""

    def __init__(self, user_id):
        self.user_id = user_id
        self.items = []

    def add_item(self, product_id, quantity=1):
        """Add item to cart - unused method."""
        self.items.append({"product_id": product_id, "quantity": quantity})

    def clear(self):
        """Clear cart - unused method."""
        self.items = []



import ast
from pathlib import Path
from skylos.rules.base import SkylosRule
from skylos.control_flow import evaluate_static_condition

class UnreachableCodeRule(SkylosRule):
    rule_id = "SKY-U001"
    name = "Unreachable Code"

    def visit_node(self, node, context):
        findings = []
        
        if isinstance(node, ast.If):
            condition = evaluate_static_condition(node.test)
            
            if condition is False and node.body:
                findings.append({
                    "rule_id": self.rule_id,
                    "kind": "quality",
                    "severity": "MEDIUM",
                    "type": "dead_branch",
                    "name": "unreachable",
                    "simple_name": "unreachable",
                    "value": "if_false",
                    "threshold": 0,
                    "message": "Dead code: condition is always False",
                    "file": context.get("filename"),
                    "basename": Path(context.get("filename", "")).name,
                    "line": node.lineno,
                    "col": node.col_offset,
                })
            elif condition is True and node.orelse:
                findings.append({
                    "rule_id": self.rule_id,
                    "kind": "quality", 
                    "severity": "MEDIUM",
                    "type": "dead_branch",
                    "name": "unreachable",
                    "simple_name": "unreachable",
                    "value": "else_after_true",
                    "threshold": 0,
                    "message": "Dead code: else branch after condition that is always True",
                    "file": context.get("filename"),
                    "basename": Path(context.get("filename", "")).name,
                    "line": node.orelse[0].lineno,
                    "col": node.orelse[0].col_offset,
                })

        if not hasattr(node, "body") or not isinstance(node.body, list):
            if findings:
                return findings
            else:
                return None

        def block_terminates(stmts):
            for s in stmts:
                if isinstance(s, (ast.Return, ast.Raise, ast.Break, ast.Continue)):
                    return True
                
                if isinstance(s, ast.If):
                    if s.orelse and block_terminates(s.body) and block_terminates(s.orelse):
                        return True
                    
            return False

        terminated = False
        
        for stmt in node.body:
            if terminated:
                findings.append({
                    "rule_id": "SKY-U002",
                    "kind": "quality",
                    "severity": "MEDIUM",
                    "type": "statement",
                    "name": "unreachable",
                    "simple_name": "unreachable",
                    "value": "dead_logic",
                    "threshold": 0,
                    "message": "Unreachable code: statement follows return/raise/break",
                    "file": context.get("filename"),
                    "basename": Path(context.get("filename", "")).name,
                    "line": stmt.lineno,
                    "col": stmt.col_offset,
                })
                break

            if isinstance(stmt, (ast.Return, ast.Raise, ast.Break, ast.Continue)):
                terminated = True
            elif isinstance(stmt, ast.If):
                if stmt.orelse and block_terminates(stmt.body) and block_terminates(stmt.orelse):
                    terminated = True

        if findings:
            return findings
        else:
            return None



import os
from .base import BaseAdapter


class OpenAIAdapter(BaseAdapter):
    def __init__(self, model, api_key):
        super().__init__(model, api_key)
        try:
            import openai
        except ImportError:
            raise ImportError("OpenAI not found. Run `pip install openai`.")

        key = self.api_key or os.getenv("OPENAI_API_KEY")
        if not key:
            raise ValueError("No OpenAI API Key found. Set OPENAI_API_KEY or login.")

        self.client = openai.OpenAI(api_key=key)

    def complete(self, system_prompt, user_prompt):
        try:
            response = self.client.responses.create(
                model=self.model, instructions=system_prompt, input=user_prompt
            )

            return response.output_text.strip()

        except Exception as e:
            return f"OpenAI Error: {str(e)}"



import datetime
import re
import random  # Unused import


def get_current_time():
    """Get current time as ISO string - used function."""
    return datetime.datetime.now().isoformat()


def format_currency(amount, currency="USD"):
    """Format amount as currency string - used function."""
    symbols = {"USD": "$", "EUR": "€", "GBP": "£"}
    symbol = symbols.get(currency, "")
    return f"{symbol}{amount:.2f}"


def generate_random_token():
    """Generate random token - unused function."""
    import secrets

    return secrets.token_hex(16)


def validate_email(email):
    """Validate email format - unused function."""
    pattern = r"^[\w\.-]+@[\w\.-]+\.\w+$"
    return bool(re.match(pattern, email))


class DateUtils:
    """Date utility class - unused class."""

    @staticmethod
    def get_tomorrow():
        """Get tomorrow's date - unused method."""
        return datetime.datetime.now() + datetime.timedelta(days=1)

    @staticmethod
    def format_date(date, format_str="%Y-%m-%d"):
        """Format date - unused method."""
        return date.strftime(format_str)



# import ast

# class BasicBlock:
#     def __init__(self, id):
#         self.id = id
#         self.statements = []
#         self.exits = []
#         self.incoming = []

#     def add_edge(self, target_block):
#         self.exits.append(target_block)
#         target_block.incoming.append(self)

# class CFGBuilder(ast.NodeVisitor):
#     def __init__(self):
#         self.blocks = []
#         self.current_block = None
#         self.cfgs = {}

#     def _new_block(self):
#         block = BasicBlock(len(self.blocks))
#         self.blocks.append(block)
#         return block

#     def visit_FunctionDef(self, node):
#         entry_block = self._new_block()
#         self.cfgs[node.name] = entry_block
#         self.current_block = entry_block

#         for stmt in node.body:
#             self.visit(stmt)

#     def visit_If(self, node):
#         self.current_block.statements.append(node.test)

#         cond_block = self.current_block
#         then_block = self._new_block()
#         else_block = self._new_block()
#         join_block = self._new_block()

#         cond_block.add_edge(then_block)
#         cond_block.add_edge(else_block)

#         self.current_block = then_block
#         for stmt in node.body:
#             self.visit(stmt)
#         self.current_block.add_edge(join_block)

#         self.current_block = else_block
#         for stmt in node.orelse:
#             self.visit(stmt)
#         self.current_block.add_edge(join_block)

#         self.current_block = join_block

#     def generic_visit(self, node):
#         if isinstance(node, ast.stmt):
#             if self.current_block:
#                 self.current_block.statements.append(node)
#         super().generic_visit(node)



from abc import ABC, abstractmethod


class SkylosRule(ABC):
    """Base class for all Skylos rules"""

    @property
    @abstractmethod
    def rule_id(self):
        pass

    @property
    @abstractmethod
    def name(self):
        pass

    @abstractmethod
    def visit_node(self, node, context):
        pass



from tree_sitter import Language, QueryCursor
import tree_sitter_typescript as tsts

try:
    TS_LANG = Language(tsts.language_typescript())
except:
    TS_LANG = None


def scan_danger(root_node, file_path):
    findings = []
    if not TS_LANG:
        return []

    def check(pattern, cap_name, rule, sev, msg):
        try:
            query = TS_LANG.query(pattern)
            cursor = QueryCursor(query)
            captures = cursor.captures(root_node)
            nodes = captures.get(cap_name, [])

            for node in nodes:
                line = node.start_point[0] + 1
                findings.append(
                    {
                        "rule_id": rule,
                        "severity": sev,
                        "message": msg,
                        "file": str(file_path),
                        "line": line,
                        "col": 0,
                    }
                )
        except Exception:
            pass

    check(
        '(call_expression function: (identifier) @eval (#eq? @eval "eval"))',
        "eval",
        "SKY-D501",
        "CRITICAL",
        "Use of eval() detected",
    )

    check(
        '(assignment_expression left: (member_expression property: (property_identifier) @xss (#eq? @xss "innerHTML")))',
        "xss",
        "SKY-D502",
        "HIGH",
        "Unsafe innerHTML assignment",
    )

    return findings



import os

try:
    import keyring

    KEYRING_AVAILABLE = True
except ImportError:
    KEYRING_AVAILABLE = False

SERVICE_NAME = "skylos"


def save_key(provider, key):
    if not KEYRING_AVAILABLE:
        print("[warn] 'keyring' library not found. Cannot save credentials securely.")
        return

    try:
        keyring.set_password(SERVICE_NAME, provider, key)
    except Exception as e:
        print(f"[warn] Failed to save to system keyring: {e}")


def get_key(provider):
    env_var = f"{provider.upper()}_API_KEY"
    key = os.getenv(env_var)
    if key:
        return key

    if KEYRING_AVAILABLE:
        try:
            return keyring.get_password(SERVICE_NAME, provider)
        except Exception:
            return None

    return None



from tree_sitter import Language, Parser, QueryCursor
import tree_sitter_typescript as tsts
from skylos.visitor import Definition

try:
    TS_LANG = Language(tsts.language_typescript())
except Exception:
    TS_LANG = None


class TypeScriptCore:
    """
    High level wrapper around a tree-sitter TS parse tree that extracts symbol
    definitions and reference occurrences from a single source file.
    """

    def __init__(self, file_path, source_bytes):
        self.file_path = file_path
        self.source = source_bytes
        self.defs = []
        self.refs = []

        if TS_LANG:
            self.parser = Parser(TS_LANG)
            self.tree = self.parser.parse(source_bytes)
            self.root_node = self.tree.root_node
        else:
            self.tree = None
            self.root_node = None

    def _get_text(self, node):
        return self.source[node.start_byte : node.end_byte].decode("utf-8")

    def _run_query(self, pattern, capture_name):
        if not self.root_node or not TS_LANG:
            return []

        try:
            query = TS_LANG.query(pattern)
            cursor = QueryCursor(query)
            captures = cursor.captures(self.root_node)

            return captures.get(capture_name, [])

        except Exception:
            return []

    def scan(self):
        if not self.root_node:
            return

        for node in self._run_query(
            "(function_declaration name: (identifier) @def)", "def"
        ):
            self._add_def(node, "function")

        for node in self._run_query(
            "(class_declaration name: (identifier) @def)", "def"
        ):
            self._add_def(node, "class")

        for node in self._run_query(
            "(method_definition name: (property_identifier) @def)", "def"
        ):
            self._add_def(node, "method")
        for node in self._run_query(
            "(method_definition name: (identifier) @def)", "def"
        ):
            self._add_def(node, "method")

        for node in self._run_query(
            "(variable_declarator name: (identifier) @def value: (arrow_function))",
            "def",
        ):
            self._add_def(node, "function")

        ref_patterns = [
            "(call_expression function: (identifier) @ref)",
            "(new_expression constructor: (identifier) @ref)",
            "(member_expression property: (property_identifier) @ref)",
        ]

        for pattern in ref_patterns:
            for node in self._run_query(pattern, "ref"):
                name = self._get_text(node)
                self.refs.append((name, self.file_path))

    def _add_def(self, node, type_name):
        name = self._get_text(node)
        line = node.start_point[0] + 1

        is_exported = False
        try:
            p = node.parent
            if p and "export" in p.type:
                is_exported = True
        except:
            pass

        d = Definition(name, type_name, self.file_path, line)
        d.is_exported = is_exported
        self.defs.append(d)



#!/usr/bin/env python3
import ast
import sys
import json
import logging
import os
from pathlib import Path
from collections import defaultdict
from skylos.visitor import Visitor
from skylos.constants import PENALTIES, AUTO_CALLED
from skylos.visitors.test_aware import TestAwareVisitor
from skylos.rules.secrets import scan_ctx as _secrets_scan_ctx
from skylos.rules.danger.danger import scan_ctx as scan_danger
import os
import traceback
from skylos.visitors.framework_aware import (
    FrameworkAwareVisitor,
    detect_framework_usage,
)

from skylos.visitors.languages.typescript import scan_typescript_file

from skylos.config import load_config
from skylos.linter import LinterVisitor
from skylos.rules.quality.complexity import ComplexityRule
from skylos.rules.quality.nesting import NestingRule
from skylos.rules.danger.calls import DangerousCallsRule
from skylos.rules.quality.structure import ArgCountRule, FunctionLengthRule
from skylos.rules.quality.logic import (
    MutableDefaultRule,
    BareExceptRule,
    DangerousComparisonRule,
    TryBlockPatternsRule

)
from skylos.rules.quality.performance import PerformanceRule
from skylos.rules.quality.unreachable import UnreachableCodeRule

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
)
logger = logging.getLogger("Skylos")


class Skylos:
    def __init__(self):
        self.defs = {}
        self.refs = []
        self.dynamic = set()
        self.exports = defaultdict(set)

    def _module(self, root, f):
        p = list(f.relative_to(root).parts)
        if p[-1].endswith(".py"):
            p[-1] = p[-1][:-3]
        if p[-1] == "__init__":
            p.pop()
        return ".".join(p)

    def _should_exclude_file(self, file_path, root_path, exclude_folders):
        if not exclude_folders:
            return False

        try:
            rel_path = file_path.relative_to(root_path)
        except ValueError:
            return False

        path_parts = rel_path.parts
        rel_path_str = str(rel_path).replace("\\", "/")

        for exclude_folder in exclude_folders:
            exclude_normalized = exclude_folder.replace("\\", "/")

            if "*" in exclude_folder:
                for part in path_parts:
                    if part.endswith(exclude_folder.replace("*", "")):
                        return True
            elif "/" in exclude_normalized:
                if rel_path_str == exclude_normalized:
                    return True
                if rel_path_str.startswith(exclude_normalized + "/"):
                    return True
                check = "/" + rel_path_str + "/"
                if "/" + exclude_normalized + "/" in check:
                    return True
            else:
                if exclude_folder in path_parts:
                    return True

        return False

    def _get_python_files(self, path, exclude_folders=None):
        p = Path(path).resolve()

        if p.is_file():
            return [p], p.parent

        root = p
        all_files = list(p.glob("**/*.py"))

        if exclude_folders:
            filtered_files = []
            excluded_count = 0

            for file_path in all_files:
                if self._should_exclude_file(file_path, root, exclude_folders):
                    excluded_count += 1
                    continue
                filtered_files.append(file_path)

            if excluded_count > 0:
                logger.info(f"Excluded {excluded_count} files from analysis")

            return filtered_files, root

        return all_files, root

    def _mark_exports(self):
        for name, definition in self.defs.items():
            if definition.in_init and not definition.simple_name.startswith("_"):
                definition.is_exported = True

        for mod, export_names in self.exports.items():
            for name in export_names:
                for def_name, def_obj in self.defs.items():
                    if (
                        def_name.startswith(f"{mod}.")
                        and def_obj.simple_name == name
                        and def_obj.type != "import"
                    ):
                        def_obj.is_exported = True

    def _mark_refs(self):
        import_to_original = {}
        for name, def_obj in self.defs.items():
            if def_obj.type == "import":
                import_name = name.split(".")[-1]

                for def_name, orig_def in self.defs.items():
                    if (
                        orig_def.type != "import"
                        and orig_def.simple_name == import_name
                        and def_name != name
                    ):
                        import_to_original[name] = def_name
                        break

        simple_name_lookup = defaultdict(list)
        for definition in self.defs.values():
            simple_name_lookup[definition.simple_name].append(definition)

        for ref, _ in self.refs:
            if ref in self.defs:
                self.defs[ref].references += 1
                if ref in import_to_original:
                    original = import_to_original[ref]
                    self.defs[original].references += 1
                continue

            simple = ref.split(".")[-1]
            ref_mod = ref.rsplit(".", 1)[0]
            candidates = simple_name_lookup.get(simple, [])

            if ref_mod:
                if ref_mod in ("cls", "self"):
                    cls_candidates = []
                    for d in candidates:
                        if d.type == "variable" and "." in d.name:
                            cls_candidates.append(d)

                    if cls_candidates:
                        for d in cls_candidates:
                            d.references += 1
                        continue

                else:
                    filtered = []
                    for d in candidates:
                        if d.name.startswith(ref_mod + ".") and d.type != "import":
                            filtered.append(d)
                    candidates = filtered
            else:
                filtered = []
                for d in candidates:
                    if d.type != "import":
                        filtered.append(d)
                candidates = filtered

            if len(candidates) == 1:
                candidates[0].references += 1
                continue

            non_import_defs = []
            for d in simple_name_lookup.get(simple, []):
                if d.type != "import":
                    non_import_defs.append(d)

            if len(non_import_defs) == 1:
                non_import_defs[0].references += 1
                continue
        
        if hasattr(self, 'pattern_trackers'):
            for tracker in self.pattern_trackers.values():
                for def_obj in self.defs.values():
                    should_mark, conf, reason = tracker.should_mark_as_used(def_obj)
                    if should_mark:
                        def_obj.references += 1

    def _get_base_classes(self, class_name):
        if class_name not in self.defs:
            return []

        class_def = self.defs[class_name]

        if hasattr(class_def, "base_classes"):
            return class_def.base_classes

        return []

    def _apply_penalties(self, def_obj, visitor, framework):
        confidence = 100

        if (
            getattr(visitor, "ignore_lines", None)
            and def_obj.line in visitor.ignore_lines
        ):
            def_obj.confidence = 0
            return

        if "." in def_obj.name:
            owner, attr = def_obj.name.rsplit(".", 1)
            owner_simple = owner.split(".")[-1]

            if (
                owner_simple == "Settings"
                or owner_simple == "Config"
                or owner_simple.endswith("Settings")
                or owner_simple.endswith("Config")
            ):
                if attr.isupper() or not attr.startswith("_"):
                    def_obj.confidence = 0
                    return

        if def_obj.type == "variable" and def_obj.simple_name == "_":
            def_obj.confidence = 0
            return

        if def_obj.simple_name.startswith("_") and not def_obj.simple_name.startswith(
            "__"
        ):
            confidence -= PENALTIES["private_name"]

        if def_obj.simple_name.startswith("__") and def_obj.simple_name.endswith("__"):
            confidence -= PENALTIES["dunder_or_magic"]

        if def_obj.in_init and def_obj.type in ("function", "class"):
            confidence -= PENALTIES["in_init_file"]

        if def_obj.name.split(".")[0] in self.dynamic:
            confidence -= PENALTIES["dynamic_module"]

        if visitor.is_test_file or def_obj.line in visitor.test_decorated_lines:
            confidence -= PENALTIES["test_related"]

        if def_obj.type == "variable" and getattr(framework, "dataclass_fields", None):
            if def_obj.name in framework.dataclass_fields:
                def_obj.confidence = 0
                return

        if def_obj.type == "variable" and "." in def_obj.name:
            prefix, _ = def_obj.name.rsplit(".", 1)

            cls_def = self.defs.get(prefix)
            if cls_def and cls_def.type == "class":
                cls_simple = cls_def.simple_name

                if (
                    getattr(framework, "pydantic_models", None)
                    and cls_simple in framework.pydantic_models
                ):
                    def_obj.confidence = 0
                    return

                cls_node = getattr(framework, "class_defs", {}).get(cls_simple)
                if cls_node is not None:
                    schema_like = False

                    for base in cls_node.bases:
                        if isinstance(base, ast.Name) and base.id.lower().endswith(
                            ("schema", "model")
                        ):
                            schema_like = True
                            break

                        if isinstance(
                            base, ast.Attribute
                        ) and base.attr.lower().endswith(("schema", "model")):
                            schema_like = True
                            break

                    if schema_like:
                        def_obj.confidence = 0
                        return

        if def_obj.type == "variable":
            fr = getattr(framework, "first_read_lineno", {}).get(def_obj.name)
            if fr is not None and fr >= def_obj.line:
                def_obj.confidence = 0
                return

        if def_obj.type == "variable" and "." in def_obj.name:
            _, attr = def_obj.name.rsplit(".", 1)

            for other in self.defs.values():
                if other is def_obj:
                    continue
                if other.type != "variable":
                    continue
                if "." not in other.name:
                    continue
                if other.simple_name != attr:
                    continue

                def_obj.confidence = 0
                return

        framework_confidence = detect_framework_usage(def_obj, visitor=framework)
        if framework_confidence is not None:
            confidence = min(confidence, framework_confidence)

        if def_obj.simple_name.startswith("__") and def_obj.simple_name.endswith("__"):
            confidence = 0

        if def_obj.type == "parameter":
            if def_obj.simple_name in ("self", "cls"):
                confidence = 0
            elif "." in def_obj.name:
                method_name = def_obj.name.split(".")[-2]
                if method_name.startswith("__") and method_name.endswith("__"):
                    confidence = 0

        if visitor.is_test_file or def_obj.line in visitor.test_decorated_lines:
            confidence = 0

        if (
            def_obj.type == "import"
            and def_obj.name.startswith("__future__.")
            and def_obj.simple_name
            in (
                "annotations",
                "absolute_import",
                "division",
                "print_function",
                "unicode_literals",
                "generator_stop",
            )
        ):
            confidence = 0

        def_obj.confidence = max(confidence, 0)

    def _apply_heuristics(self):
        class_methods = defaultdict(list)
        for definition in self.defs.values():
            if definition.type in ("method", "function") and "." in definition.name:
                cls = definition.name.rsplit(".", 1)[0]
                if cls in self.defs and self.defs[cls].type == "class":
                    class_methods[cls].append(definition)

        for cls, methods in class_methods.items():
            if self.defs[cls].references > 0:
                for method in methods:
                    if method.simple_name in AUTO_CALLED:
                        method.references += 1

                    if (
                        method.simple_name.startswith("visit_")
                        or method.simple_name.startswith("leave_")
                        or method.simple_name.startswith("transform_")
                    ):
                        method.references += 1

                    if method.simple_name == "format" and cls.endswith("Formatter"):
                        method.references += 1

    def analyze(
        self,
        path,
        thr=60,
        exclude_folders=None,
        enable_secrets=False,
        enable_danger=False,
        enable_quality=False,
        extra_visitors=None,
    ):
        files, root = self._get_python_files(path, exclude_folders)

        if not files:
            logger.warning(f"No Python files found in {path}")
            return json.dumps(
                {
                    "unused_functions": [],
                    "unused_imports": [],
                    "unused_classes": [],
                    "unused_variables": [],
                    "unused_parameters": [],
                    "analysis_summary": {
                        "total_files": 0,
                        "excluded_folders": exclude_folders if exclude_folders else [],
                    },
                }
            )

        logger.info(f"Analyzing {len(files)} Python files...")

        modmap = {}
        for f in files:
            modmap[f] = self._module(root, f)

        from skylos.implicit_refs import pattern_tracker
        if Path(".coverage").exists():
            if pattern_tracker.load_coverage():
                logger.info(f"Loaded coverage data ({len(pattern_tracker.coverage_hits)} lines)")

        all_secrets = []
        all_dangers = []
        all_quality = []
        file_contexts = []
        
        pattern_trackers = {} 

        for file in files:
            mod = modmap[file]
            (
                defs,
                refs,
                dyn,
                exports,
                test_flags,
                framework_flags,
                q_finds,
                d_finds,
                pro_finds,
                pattern_tracker
            ) = proc_file(file, mod, extra_visitors)

            if pattern_tracker:
                pattern_trackers[mod] = pattern_tracker

            for definition in defs:
                self.defs[definition.name] = definition

            self.refs.extend(refs)
            self.dynamic.update(dyn)
            self.exports[mod].update(exports)

            file_contexts.append((defs, test_flags, framework_flags, file, mod))


            if enable_quality and q_finds:
                all_quality.extend(q_finds)

            if enable_danger and d_finds:
                all_dangers.extend(d_finds)

            if pro_finds:
                all_dangers.extend(pro_finds)

            if enable_secrets and _secrets_scan_ctx is not None:
                try:
                    src = Path(file).read_text(encoding="utf-8", errors="ignore")
                    src_lines = src.splitlines(True)
                    rel = str(Path(file).relative_to(root))
                    ctx = {"relpath": rel, "lines": src_lines, "tree": None}
                    findings = list(_secrets_scan_ctx(ctx))
                    if findings:
                        all_secrets.extend(findings)
                except Exception:
                    pass
        
        self.pattern_trackers = pattern_trackers 

        for defs, test_flags, framework_flags, file, mod in file_contexts:
            for definition in defs:
                self._apply_penalties(definition, test_flags, framework_flags)

            if enable_danger and scan_danger is not None:
                try:
                    findings = scan_danger(root, [file])
                    if findings:
                        all_dangers.extend(findings)
                except Exception as e:
                    logger.error(f"Error scanning {file} for dangerous code: {e}")
                    if os.getenv("SKYLOS_DEBUG"):
                        logger.error(traceback.format_exc())

        self._mark_refs()
        self._apply_heuristics()
        self._mark_exports()

        shown = 0

        def def_sort_key(d):
            return (d.type, d.name)

        for d in sorted(self.defs.values(), key=def_sort_key):
            if shown >= 50:
                break
            shown += 1

        unused = []
        for definition in self.defs.values():
            if (
                definition.references == 0
                and not definition.is_exported
                and definition.confidence > 0
                and definition.confidence >= thr
            ):
                unused.append(definition.to_dict())

        context_map = {}
        for name, d in self.defs.items():
            if d.type in ("class", "function", "method") and not name.startswith("_"):
                context_map[name] = {
                    "name": d.name,
                    "file": str(d.filename),
                    "line": d.line,
                    "type": d.type,
                }
        result = {
            "definitions": context_map,
            "unused_functions": [],
            "unused_imports": [],
            "unused_classes": [],
            "unused_variables": [],
            "unused_parameters": [],
            "analysis_summary": {
                "total_files": len(files),
                "excluded_folders": exclude_folders or [],
            },
        }

        if enable_secrets and all_secrets:
            result["secrets"] = all_secrets
            result["analysis_summary"]["secrets_count"] = len(all_secrets)

        if enable_danger and all_dangers:
            result["danger"] = all_dangers
            result["analysis_summary"]["danger_count"] = len(all_dangers)

        if enable_quality and all_quality:
            result["quality"] = all_quality
            result["analysis_summary"]["quality_count"] = len(all_quality)

        for u in unused:
            if u["type"] in ("function", "method"):
                result["unused_functions"].append(u)
            elif u["type"] == "import":
                result["unused_imports"].append(u)
            elif u["type"] == "class":
                result["unused_classes"].append(u)
            elif u["type"] == "variable":
                result["unused_variables"].append(u)
            elif u["type"] == "parameter":
                result["unused_parameters"].append(u)

        return json.dumps(result, indent=2)


def proc_file(file_or_args, mod=None, extra_visitors=None):
    if mod is None and isinstance(file_or_args, tuple):
        file, mod = file_or_args
    else:
        file = file_or_args

    cfg = load_config(file)

    if str(file).endswith((".ts", ".tsx")):
        return scan_typescript_file(file)

    try:
        source = Path(file).read_text(encoding="utf-8")
        ignore_lines = {
            i
            for i, line in enumerate(source.splitlines(), start=1)
            if "pragma: no skylos" in line
        }
        tree = ast.parse(source)

        q_rules = []
        if "SKY-Q301" not in cfg["ignore"]:
            q_rules.append(ComplexityRule(threshold=cfg["complexity"]))
        if "SKY-Q302" not in cfg["ignore"]:
            q_rules.append(NestingRule(threshold=cfg["nesting"]))
        if "SKY-C303" not in cfg["ignore"]:
            q_rules.append(ArgCountRule(max_args=cfg["max_args"]))
        if "SKY-C304" not in cfg["ignore"]:
            q_rules.append(FunctionLengthRule(max_lines=cfg["max_lines"]))

        if "SKY-L001" not in cfg["ignore"]:
            q_rules.append(MutableDefaultRule())
        if "SKY-L002" not in cfg["ignore"]:
            q_rules.append(BareExceptRule())
        if "SKY-L003" not in cfg["ignore"]:
            q_rules.append(DangerousComparisonRule())
        if "SKY-L004" not in cfg["ignore"]:
            q_rules.append(TryBlockPatternsRule(max_lines=15))
        
        if "SKY-U001" not in cfg["ignore"]:
            q_rules.append(UnreachableCodeRule())

        q_rules.append(PerformanceRule(ignore_list=cfg["ignore"]))

        linter_q = LinterVisitor(q_rules, str(file))
        linter_q.visit(tree)
        quality_findings = linter_q.findings

        d_rules = [DangerousCallsRule()]
        linter_d = LinterVisitor(d_rules, str(file))
        linter_d.visit(tree)
        danger_findings = linter_d.findings

        pro_findings = []
        if extra_visitors:
            for VisitorClass in extra_visitors:
                checker = VisitorClass(file, pro_findings)
                checker.visit(tree)

        tv = TestAwareVisitor(filename=file)
        tv.visit(tree)
        tv.ignore_lines = ignore_lines

        fv = FrameworkAwareVisitor(filename=file)
        fv.visit(tree)
        fv.finalize()
        v = Visitor(mod, file)
        v.visit(tree)

        fv.dataclass_fields = getattr(v, "dataclass_fields", set())
        fv.first_read_lineno = getattr(v, "first_read_lineno", {})

        return (
            v.defs,
            v.refs,
            v.dyn,
            v.exports,
            tv,
            fv,
            quality_findings,
            danger_findings,
            pro_findings,
            v.pattern_tracker
        )

    except Exception as e:
        logger.error(f"{file}: {e}")
        if os.getenv("SKYLOS_DEBUG"):
            logger.error(traceback.format_exc())
        dummy_visitor = TestAwareVisitor(filename=file)
        dummy_visitor.ignore_lines = set()
        dummy_framework_visitor = FrameworkAwareVisitor(filename=file)
        return [], [], set(), set(), dummy_visitor, dummy_framework_visitor, [], [], [], None


def analyze(
    path,
    conf=60,
    exclude_folders=None,
    enable_secrets=False,
    enable_danger=False,
    enable_quality=False,
    extra_visitors=None,
):
    return Skylos().analyze(
        path,
        conf,
        exclude_folders,
        enable_secrets,
        enable_danger,
        enable_quality,
        extra_visitors,
    )


if __name__ == "__main__":
    enable_secrets = "--secrets" in sys.argv
    enable_danger = "--danger" in sys.argv
    enable_quality = "--quality" in sys.argv

    positional = [a for a in sys.argv[1:] if not a.startswith("--")]
    if not positional:
        print(
            "Usage: python Skylos.py <path> [confidence_threshold] [--secrets] [--danger] [--quality]"
        )
        sys.exit(2)
    p = positional[0]
    confidence = int(positional[1]) if len(positional) > 1 else 60

    result = analyze(
        p,
        confidence,
        enable_secrets=enable_secrets,
        enable_danger=enable_danger,
        enable_quality=enable_quality,
    )
    data = json.loads(result)
    print("\n Python Static Analysis Results")
    print("===================================\n")

    total_dead = 0
    for key, items in data.items():
        if key.startswith("unused_") and isinstance(items, list):
            total_dead += len(items)

    danger_count = (
        data.get("analysis_summary", {}).get("danger_count", 0) if enable_danger else 0
    )
    secrets_count = (
        data.get("analysis_summary", {}).get("secrets_count", 0)
        if enable_secrets
        else 0
    )

    print("Summary:")
    if data["unused_functions"]:
        print(f" * Unreachable functions: {len(data['unused_functions'])}")
    if data["unused_imports"]:
        print(f" * Unused imports: {len(data['unused_imports'])}")
    if data["unused_classes"]:
        print(f" * Unused classes: {len(data['unused_classes'])}")
    if data["unused_variables"]:
        print(f" * Unused variables: {len(data['unused_variables'])}")
    if enable_danger:
        print(f" * Security issues: {danger_count}")
    if enable_secrets:
        print(f" * Secrets found: {secrets_count}")

    if data["unused_functions"]:
        print("\n - Unreachable Functions")
        print("=======================")
        for i, func in enumerate(data["unused_functions"], 1):
            print(f" {i}. {func['name']}")
            print(f"    └─ {func['file']}:{func['line']}")

    if data["unused_imports"]:
        print("\n - Unused Imports")
        print("================")
        for i, imp in enumerate(data["unused_imports"], 1):
            print(f" {i}. {imp['simple_name']}")
            print(f"    └─ {imp['file']}:{imp['line']}")

    if data["unused_classes"]:
        print("\n - Unused Classes")
        print("=================")
        for i, cls in enumerate(data["unused_classes"], 1):
            print(f" {i}. {cls['name']}")
            print(f"    └─ {cls['file']}:{cls['line']}")

    if data["unused_variables"]:
        print("\n - Unused Variables")
        print("==================")
        for i, var in enumerate(data["unused_variables"], 1):
            print(f" {i}. {var['name']}")
            print(f"    └─ {var['file']}:{var['line']}")

    if enable_danger and data.get("danger"):
        print("\n - Security Issues")
        print("================")
        for i, f in enumerate(data["danger"], 1):
            print(
                f" {i}. {f['message']} [{f['rule_id']}] ({f['file']}:{f['line']}) Severity: {f['severity']}"
            )

    if enable_secrets and data.get("secrets"):
        print("\n - Secrets")
        print("==========")
        for i, s in enumerate(data["secrets"], 1):
            rid = s.get("rule_id", "SECRET")
            msg = s.get("message", "Potential secret")
            file = s.get("file")
            line = s.get("line", 1)
            sev = s.get("severity", "HIGH")
            print(f" {i}. {msg} [{rid}] ({file}:{line}) Severity: {sev}")

    print("\n" + "─" * 50)
    if enable_danger:
        print(
            f"Found {total_dead} dead code items and {danger_count} security flaws. Add this badge to your README:"
        )
    else:
        print(f"Found {total_dead} dead code items. Add this badge to your README:")
    print("```markdown")
    print(
        f"![Dead Code: {total_dead}](https://img.shields.io/badge/Dead_Code-{total_dead}_detected-orange?logo=codacy&logoColor=red)"
    )
    print("```")

    print("\nNext steps:")
    print("  * Use --interactive to select specific items to remove")
    print("  * Use --dry-run to preview changes before applying them")



import unittest
from unittest.mock import MagicMock, patch
from skylos.fixer import Fixer


class TestFixer(unittest.TestCase):
    def setUp(self):
        self.mock_defs = {
            "used_function": {"name": "used_function", "type": "function", "line": 10},
            "unused_class": {"name": "UnusedClass", "type": "class", "line": 20},
        }
        self.source_code = """
import os
def main():
    used_function()
"""

    @patch("skylos.fixer.get_adapter")
    def test_initialization(self, mock_get_adapter):
        mock_adapter_instance = MagicMock()
        mock_get_adapter.return_value = mock_adapter_instance

        fixer = Fixer(api_key="sk-test", model="gpt-4.1")

        mock_get_adapter.assert_called_with("gpt-4.1", "sk-test")
        self.assertIsNotNone(fixer.adapter)

    def test_line_number_injection(self):
        with patch("skylos.fixer.get_adapter"):
            fixer = Fixer(api_key="test")

            code = "line1\nline2"
            numbered = fixer._add_line_numbers(code)

            expected = "   1 | line1\n   2 | line2"
            self.assertEqual(numbered, expected)

    def test_context_retrieval(self):
        with patch("skylos.fixer.get_adapter"):
            fixer = Fixer(api_key="test")

            context = fixer._get_relevant_context(self.source_code, self.mock_defs)

            self.assertIn("used_function", context)
            self.assertNotIn("UnusedClass", context)

    @patch("skylos.fixer.get_adapter")
    def test_audit_file_flow(self, mock_get_adapter):
        mock_adapter = MagicMock()
        mock_adapter.complete.return_value = "[SECURITY] Line 10: Bad stuff"
        mock_get_adapter.return_value = mock_adapter

        fixer = Fixer(api_key="test")
        result = fixer.audit_file(self.source_code, self.mock_defs)

        self.assertEqual(result, "[SECURITY] Line 10: Bad stuff")

        call_args = mock_adapter.complete.call_args
        sent_prompt = call_args[0][1]
        self.assertIn("   1 |", sent_prompt)
        self.assertIn("used_function", sent_prompt)

    @patch("skylos.fixer.get_adapter")
    def test_fix_bug_parsing(self, mock_get_adapter):
        mock_adapter = MagicMock()
        mock_adapter.complete.return_value = """
---PROBLEM---
It was broken
---CHANGE---
Fixed it
---CODE---
def main():
    print("fixed")
"""
        mock_get_adapter.return_value = mock_adapter

        fixer = Fixer(api_key="test")
        result = fixer.fix_bug("bad_code", 1, "Error", self.mock_defs)

        self.assertEqual(result["problem"], "It was broken")
        self.assertEqual(result["change"], "Fixed it")
        self.assertEqual(result["code"], 'def main():\n    print("fixed")')

    @patch("skylos.fixer.get_adapter")
    def test_fix_bug_parsing_fallback(self, mock_get_adapter):
        mock_adapter = MagicMock()
        mock_adapter.complete.return_value = "print('just code')"
        mock_get_adapter.return_value = mock_adapter

        fixer = Fixer(api_key="test")
        result = fixer.fix_bug("bad_code", 1, "Error", self.mock_defs)

        self.assertEqual(result["problem"], "Issue detected")
        self.assertEqual(result["code"], "print('just code')")


if __name__ == "__main__":
    unittest.main()



import os
import sys
import argparse
import csv  # Unused import
from functools import wraps

from sample_repo import User, Product
from sample_repo import get_current_time


def register_commands():
    """Register CLI commands - used function."""
    parser = argparse.ArgumentParser(description="Sample CLI")
    subparsers = parser.add_subparsers(dest="command")

    user_parser = subparsers.add_parser("user", help="User commands")
    user_parser.add_argument("action", choices=["list", "create"])

    product_parser = subparsers.add_parser("product", help="Product commands")
    product_parser.add_argument("action", choices=["list"])

    return parser


def handle_user_command(args):
    """Handle user command - used function."""
    if args.action == "list":
        users = [User("john", "john@example.com"), User("alice", "alice@example.com")]
        for user in users:
            print(f"User: {user.username} ({user.email})")
    elif args.action == "create":
        user = User("new_user", "new@example.com")
        print(f"Created user: {user.username} ({user.email})")


def handle_product_command(args):
    """Handle product command - used function."""
    if args.action == "list":
        products = [Product("Laptop", 999.99), Product("Phone", 499.99)]
        for product in products:
            print(f"Product: {product.name} - ${product.price:.2f}")


def export_data(data, filename):
    """Export data to CSV - unused function."""
    with open(filename, "w", newline="") as f:
        writer = csv.writer(f)
        writer.writerow(["id", "name", "value"])
        for item in data:
            writer.writerow([item["id"], item["name"], item["value"]])


def import_data(filename):
    """Import data from CSV - unused function."""
    data = []
    with open(filename, "r", newline="") as f:
        reader = csv.reader(f)
        next(reader)
        for row in reader:
            data.append({"id": row[0], "name": row[1], "value": row[2]})
    return data


class CommandDecorator:
    """Decorator for CLI commands - unused class."""

    def __init__(self, name, help_text):
        self.name = name
        self.help_text = help_text

    def __call__(self, func):
        """Call method - unused method."""

        @wraps(func)
        def wrapper(*args, **kwargs):
            print(f"Executing command: {self.name}")
            return func(*args, **kwargs)

        return wrapper



import ast
from skylos.rules.base import SkylosRule
from pathlib import Path


class ArgCountRule(SkylosRule):
    rule_id = "SKY-C303"
    name = "Too Many Arguments"

    def __init__(self, max_args=5):
        self.max_args = max_args

    def visit_node(self, node, context):
        if not isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
            return None

        args = node.args.args
        clean_args = []
        for a in args:
            if a.arg not in ("self", "cls"):
                clean_args.append(a)

        total_count = len(clean_args) + len(node.args.kwonlyargs)

        if total_count <= self.max_args:
            return None

        mod = context.get("mod", "")
        if mod:
            func_name = f"{mod}.{node.name}"
        else:
            func_name = node.name

        return [
            {
                "rule_id": self.rule_id,
                "kind": "structure",
                "type": "function",
                "name": func_name,
                "simple_name": node.name,
                "value": total_count,
                "threshold": self.max_args,
                "severity": "MEDIUM",
                "message": f"Function has {total_count} arguments (limit: {self.max_args}). Refactor.",
                "file": context.get("filename"),
                "basename": Path(context.get("filename", "")).name,
                "line": node.lineno,
                "col": node.col_offset,
            }
        ]


class FunctionLengthRule(SkylosRule):
    rule_id = "SKY-C304"
    name = "Function Too Long"

    def __init__(self, max_lines=50):
        self.max_lines = max_lines

    def visit_node(self, node, context):
        if not isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
            return None

        start = getattr(node, "lineno", 0)
        end = getattr(node, "end_lineno", start)
        physical_length = max(end - start + 1, 0)

        if physical_length <= self.max_lines:
            return None

        if physical_length < 100:
            severity = "MEDIUM"
        else:
            severity = "HIGH"

        mod = context.get("mod", "")

        if mod:
            func_name = f"{mod}.{node.name}"
        else:
            func_name = node.name

        return [
            {
                "rule_id": self.rule_id,
                "kind": "structure",
                "type": "function",
                "name": func_name,
                "simple_name": node.name,
                "value": physical_length,
                "threshold": self.max_lines,
                "severity": severity,
                "message": f"Function is {physical_length} lines long (limit: {self.max_lines}). It is too big.",
                "file": context.get("filename"),
                "basename": Path(context.get("filename", "")).name,
                "line": node.lineno,
                "col": node.col_offset,
            }
        ]



import ast
from pathlib import Path
from skylos.rules.base import SkylosRule


class MutableDefaultRule(SkylosRule):
    rule_id = "SKY-L001"
    name = "Mutable Default Argument"

    def visit_node(self, node, context):
        if not isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
            return None
        findings = []

        kw_defaults_filtered = []
        for d in node.args.kw_defaults:
            if d:
                kw_defaults_filtered.append(d)

        for default in node.args.defaults + kw_defaults_filtered:
            if isinstance(default, (ast.List, ast.Dict, ast.Set)):
                findings.append(
                    {
                        "rule_id": self.rule_id,
                        "kind": "logic",
                        "severity": "HIGH",
                        "type": "function",
                        "name": node.name,
                        "simple_name": node.name,
                        "value": "mutable",
                        "threshold": 0,
                        "message": "Mutable default argument detected (List/Dict/Set). This causes state leaks.",
                        "file": context.get("filename"),
                        "basename": Path(context.get("filename", "")).name,
                        "line": default.lineno,
                        "col": default.col_offset,
                    }
                )
        if findings:
            return findings
        else:
            return None


class BareExceptRule(SkylosRule):
    rule_id = "SKY-L002"
    name = "Bare Except Block"

    def visit_node(self, node, context):
        if isinstance(node, ast.ExceptHandler) and node.type is None:
            return [
                {
                    "rule_id": self.rule_id,
                    "kind": "logic",
                    "severity": "MEDIUM",
                    "type": "block",
                    "name": "except",
                    "simple_name": "except",
                    "value": "bare",
                    "threshold": 0,
                    "message": "Bare 'except:' block swallows SystemExit and other critical errors.",
                    "file": context.get("filename"),
                    "basename": Path(context.get("filename", "")).name,
                    "line": node.lineno,
                    "col": node.col_offset,
                }
            ]
        return None


class DangerousComparisonRule(SkylosRule):
    rule_id = "SKY-L003"
    name = "Dangerous Comparison"

    def visit_node(self, node, context):
        if not isinstance(node, ast.Compare):
            return None

        findings = []
        for op, comparator in zip(node.ops, node.comparators):
            if isinstance(op, (ast.Eq, ast.NotEq)):
                if isinstance(comparator, ast.Constant):
                    val = comparator.value
                    if val is True or val is False or val is None:
                        findings.append(
                            {
                                "rule_id": self.rule_id,
                                "kind": "logic",
                                "severity": "LOW",
                                "type": "comparison",
                                "name": "==",
                                "simple_name": "==",
                                "value": str(comparator.value),
                                "threshold": 0,
                                "message": f"Comparison to {comparator.value} should use 'is' or 'is not'.",
                                "file": context.get("filename"),
                                "basename": Path(context.get("filename", "")).name,
                                "line": node.lineno,
                                "col": node.col_offset,
                            }
                        )
        if findings:
            return findings
        else:
            return None
        
class TryBlockPatternsRule(SkylosRule):
    rule_id = "SKY-L004"
    name = "Anti-Pattern Try Block"

    def __init__(self, max_lines=15, max_control_flow=3):
        self.max_lines = max_lines
        self.max_control_flow = max_control_flow

    def visit_node(self, node, context):
        if not isinstance(node, ast.Try):
            return None

        findings = []
  
        if node.body:
            start = node.body[0].lineno
            end = getattr(node.body[-1], "end_lineno", start)
            length = end - start + 1

            if length > self.max_lines:
                findings.append(self._create_finding(
                    node, context, 
                    severity="LOW",
                    value=length,
                    msg=f"Try block covers {length} lines (limit: {self.max_lines}). Reduce scope to the risky operation only."
                ))


        control_flow_count = 0
        has_nested_try = False

        for stmt in node.body:
            for child in ast.walk(stmt):
                if isinstance(child, ast.Try):
                    has_nested_try = True
                
                if isinstance(child, (ast.If, ast.For, ast.While)):
                    control_flow_count += 1

        if has_nested_try:
            findings.append(self._create_finding(
                node, context,
                severity="MEDIUM",
                value="nested",
                msg="Nested 'try' block detected. Flatten logic or move inner try to a helper function."
            ))

        if control_flow_count > self.max_control_flow:
            findings.append(self._create_finding(
                node, context,
                severity="HIGH",
                value=control_flow_count,
                msg=f"Try block contains {control_flow_count} control flow statements. Don't wrap complex logic in error handling."
            ))

        if findings:
            return findings
        else: 
            return None

    def _create_finding(self, node, context, severity, value, msg):
        return {
            "rule_id": self.rule_id,
            "kind": "quality",
            "severity": severity,
            "type": "block",
            "name": "try",
            "simple_name": "try",
            "value": value,
            "threshold": 0,
            "message": msg,
            "file": context.get("filename"),
            "basename": Path(context.get("filename", "")).name,
            "line": node.lineno,
            "col": node.col_offset,
        }


import ast
import operator

OPS = {
    ast.Eq: operator.eq,
    ast.NotEq: operator.ne,
    ast.Lt: operator.lt,
    ast.LtE: operator.le,
    ast.Gt: operator.gt,
    ast.GtE: operator.ge,
    ast.Is: operator.is_,
    ast.IsNot: operator.is_not,
    ast.In: lambda x, y: x in y,
    ast.NotIn: lambda x, y: x not in y,
}

def evaluate_static_condition(node):

    if isinstance(node, ast.Constant):
        return node.value

    if isinstance(node, ast.UnaryOp) and isinstance(node.op, ast.Not):
        val = evaluate_static_condition(node.operand)
        if val is not None:
            return not val
        else:
            return None

    if isinstance(node, ast.BoolOp):
        values = []
        for v in node.values:
            values.append(evaluate_static_condition(v))
        
        if isinstance(node.op, ast.And):
            for v in values:
                if v is False:
                    return False

            for v in values:
                if v is None:
                    return None

            return True

        if isinstance(node.op, ast.Or):
            for v in values:
                if v is True:
                    return True

            for v in values:
                if v is None:
                    return None

            return False

    if isinstance(node, ast.Compare):
        if len(node.ops) == 1 and len(node.comparators) == 1:
            left = evaluate_static_condition(node.left)
            right = evaluate_static_condition(node.comparators[0])
            op_type = type(node.ops[0])

            if left is not None and right is not None and op_type in OPS:
                try:
                    return OPS[op_type](left, right)
                except Exception:
                    return None

    return None

def extract_constant_string(node):
    if isinstance(node, ast.Constant) and isinstance(node.value, str):
        return node.value
    return None


import os
from .base import BaseAdapter


class AnthropicAdapter(BaseAdapter):
    def __init__(self, model: str, api_key):
        super().__init__(model, api_key)
        try:
            import anthropic
        except ImportError:
            raise ImportError("Anthropic SDK not found. Run `pip install anthropic`.")

        key = self.api_key or os.getenv("ANTHROPIC_API_KEY")
        if not key:
            raise ValueError("No Anthropic API Key found. Set ANTHROPIC_API_KEY.")

        self.client = anthropic.Anthropic(api_key=key)

    def complete(self, system_prompt, user_prompt):
        try:
            response = self.client.messages.create(
                model=self.model,
                max_tokens=4096,
                system=system_prompt,
                messages=[{"role": "user", "content": user_prompt}],
                temperature=0.2,
            )
            return response.content[0].text
        except Exception as e:
            return f"Anthropic Error: {str(e)}"



import os
import requests
import subprocess
from skylos.credentials import get_key

API_URL = os.getenv("SKYLOS_API_URL", "http://localhost:3000/api")


def get_project_token():
    return os.getenv("SKYLOS_TOKEN") or get_key("skylos_token")


def get_git_root():
    try:
        return (
            subprocess.check_output(
                ["git", "rev-parse", "--show-toplevel"], stderr=subprocess.DEVNULL
            )
            .decode()
            .strip()
        )
    except:
        return None


def get_git_info():
    try:
        commit = (
            subprocess.check_output(
                ["git", "rev-parse", "HEAD"], stderr=subprocess.DEVNULL
            )
            .decode()
            .strip()
        )
        branch = (
            subprocess.check_output(
                ["git", "rev-parse", "--abbrev-ref", "HEAD"], stderr=subprocess.DEVNULL
            )
            .decode()
            .strip()
        )
        try:
            actor = (
                subprocess.check_output(
                    ["git", "config", "user.email"], stderr=subprocess.DEVNULL
                )
                .decode()
                .strip()
            )
        except:
            actor = os.getenv("USER", "unknown")
        return commit, branch, actor
    except Exception:
        return "unknown", "unknown", "unknown"


def extract_snippet(file_path, line_number, context=3):
    try:
        with open(file_path, "r", encoding="utf-8", errors="ignore") as f:
            lines = f.readlines()
        start = max(0, line_number - 1 - context)
        end = min(len(lines), line_number + context)
        snippet_lines = [line.rstrip() for line in lines[start:end]]
        return "\n".join(snippet_lines)
    except Exception:
        return None


def check_scan_status(scan_id):
    token = get_project_token()
    try:
        response = requests.get(
            f"{API_URL}/scans/{scan_id}/poll",
            headers={"Authorization": f"Bearer {token}"},
            timeout=5,
        )
        if response.status_code == 200:
            return response.json()
        return None
    except (requests.RequestException, ValueError):
        return None


def upload_report(result_json):
    token = get_project_token()
    if not token:
        return {"success": False, "error": "No token found"}

    commit, branch, actor = get_git_info()

    git_root = get_git_root()

    def process_findings(items, category):
        processed = []
        for item in items:
            file_abs = os.path.abspath(item["file"])
            snippet = extract_snippet(file_abs, item["line"])

            file_rel = item["file"]
            if git_root:
                try:
                    file_rel = os.path.relpath(file_abs, git_root)
                except:
                    file_rel = item["file"]

            processed.append(
                {
                    "rule_id": item.get("rule_id", "UNKNOWN"),
                    "file_path": file_rel,
                    "line_number": item["line"],
                    "message": item.get("message", f"Unused {item.get('name')}"),
                    "severity": item.get("severity", "MEDIUM"),
                    "category": category,
                    "snippet": snippet,
                }
            )
        return processed

    all_findings = []
    all_findings.extend(process_findings(result_json.get("danger", []), "SECURITY"))
    all_findings.extend(process_findings(result_json.get("quality", []), "QUALITY"))
    all_findings.extend(process_findings(result_json.get("secrets", []), "SECRET"))

    unused_raw = (
        result_json.get("unused_functions", [])
        + result_json.get("unused_imports", [])
        + result_json.get("unused_variables", [])
        + result_json.get("unused_classes", [])
    )
    all_findings.extend(process_findings(unused_raw, "DEAD_CODE"))

    payload = {
        "summary": result_json.get("analysis_summary", {}),
        "findings": all_findings,
        "commit_hash": commit,
        "branch": branch,
        "actor": actor,
        "is_forced": True,
    }

    try:
        response = requests.post(
            f"{API_URL}/report",
            json=payload,
            headers={
                "Authorization": f"Bearer {token}",
                "Content-Type": "application/json",
            },
            timeout=10,
        )

        if response.status_code in (200, 201):
            data = response.json()
            return {
                "success": True,
                "scan_id": data.get("scanId"),
                "quality_gate": data.get("quality_gate"),
            }
        else:
            return {
                "success": False,
                "error": f"Server Error ({response.status_code}): {response.text}",
            }

    except Exception as e:
        return {"success": False, "error": f"Connection Error: {str(e)}"}



#!/usr/bin/env python3
import pytest
import json
import sys
import logging
from unittest.mock import Mock, patch, mock_open

from skylos.cli import (
    Colors,
    CleanFormatter,
    setup_logger,
    remove_unused_import,
    remove_unused_function,
    interactive_selection,
    print_badge,
    main,
)


class TestColors:
    def test_colors_defined(self):
        """Test that all color constants are defined."""
        assert hasattr(Colors, "RED")
        assert hasattr(Colors, "GREEN")
        assert hasattr(Colors, "RESET")
        assert hasattr(Colors, "BOLD")

        assert Colors.RED.startswith("\033[")
        assert Colors.RESET == "\033[0m"


class TestCleanFormatter:
    def test_clean_formatter_removes_metadata(self):
        """Test that CleanFormatter only returns the message."""
        formatter = CleanFormatter()

        record = Mock()
        record.getMessage.return_value = "Test message"

        result = formatter.format(record)
        assert result == "Test message"

        record.getMessage.assert_called_once()


class TestSetupLogger:
    @patch("skylos.cli.logging.FileHandler")
    @patch("skylos.cli.RichHandler")
    def test_setup_logger_console_only(self, mock_rich_handler, mock_file_handler):
        """Test logger setup without output file."""
        mock_handler = Mock()
        mock_rich_handler.return_value = mock_handler

        logger = setup_logger()

        mock_rich_handler.assert_called_once()
        mock_file_handler.assert_not_called()

    @patch("skylos.cli.logging.FileHandler")
    @patch("skylos.cli.RichHandler")
    def test_setup_logger_with_output_file(self, mock_rich_handler, mock_file_handler):
        """Test logger setup with output file."""
        mock_rich_handler.return_value = Mock()
        mock_file_handler.return_value = Mock()

        logger = setup_logger("output.log")

        mock_rich_handler.assert_called_once()
        mock_file_handler.assert_called_once_with("output.log")

    def test_remove_simple_import(self):
        """Test removing a simple import statement."""
        content = """import os
import sys
import json

def main():
    print(sys.version)
"""

        with (
            patch("pathlib.Path.read_text", return_value=content) as mock_read,
            patch("pathlib.Path.write_text") as mock_write,
            patch(
                "skylos.cli.remove_unused_import_cst", return_value=("NEW_CODE", True)
            ) as mock_codemod,
        ):
            result = remove_unused_import("test.py", "os", 1)

            assert result is True
            mock_read.assert_called_once()
            mock_codemod.assert_called_once()
            mock_write.assert_called_once_with("NEW_CODE", encoding="utf-8")

    def test_remove_from_multi_import(self):
        content = "import os, sys, json\n"

        with (
            patch("pathlib.Path.read_text", return_value=content),
            patch("pathlib.Path.write_text") as mock_write,
            patch("skylos.cli.remove_unused_import_cst", return_value=("X", True)),
        ):
            result = remove_unused_import("test.py", "os", 1)

            assert result is True
            mock_write.assert_called_once()

    def test_remove_from_import_statement(self):
        content = "from collections import defaultdict, Counter\n"

        with (
            patch("pathlib.Path.read_text", return_value=content),
            patch("pathlib.Path.write_text") as mock_write,
            patch("skylos.cli.remove_unused_import_cst", return_value=("X", True)),
        ):
            result = remove_unused_import("test.py", "Counter", 1)

            assert result is True
            mock_write.assert_called_once()

    def test_remove_entire_from_import(self):
        content = "from collections import defaultdict\n"

        with (
            patch("pathlib.Path.read_text", return_value=content),
            patch("pathlib.Path.write_text") as mock_write,
            patch("skylos.cli.remove_unused_import_cst", return_value=("", True)),
        ):
            result = remove_unused_import("test.py", "defaultdict", 1)

            assert result is True
            mock_write.assert_called_once_with("", encoding="utf-8")

    def test_remove_import_file_error(self):
        """handling file errors when removing imports."""
        with patch(
            "pathlib.Path.read_text", side_effect=FileNotFoundError("File not found")
        ):
            result = remove_unused_import("nonexistent.py", "os", 1)
            assert result is False


class TestRemoveUnusedFunction:
    def test_remove_simple_function(self):
        """test remove a simple function."""
        content = """def used_function():
    return "used"

def unused_function():
    return "unused"

def another_function():
    return "another"
"""

        with (
            patch("pathlib.Path.read_text", return_value=content),
            patch("pathlib.Path.write_text") as mock_write,
            patch(
                "skylos.cli.remove_unused_function_cst",
                return_value=("NEW_FUNC_CODE", True),
            ),
        ):
            result = remove_unused_function("test.py", "unused_function", 4)

        assert result is True
        mock_write.assert_called_once_with("NEW_FUNC_CODE", encoding="utf-8")

    def test_remove_function_with_decorators(self):
        """removing function with decorators."""
        content = """@property
@decorator
def unused_function():
    return "unused"
"""

        with (
            patch("pathlib.Path.read_text", return_value=content),
            patch("pathlib.Path.write_text") as mock_write,
            patch("skylos.cli.remove_unused_function_cst", return_value=("X", True)),
        ):
            result = remove_unused_function("test.py", "unused_function", 3)

        assert result is True
        mock_write.assert_called_once()

    def test_remove_function_file_error(self):
        with patch(
            "pathlib.Path.read_text", side_effect=FileNotFoundError("File not found")
        ):
            result = remove_unused_function("nonexistent.py", "func", 1)
            assert result is False

    def test_remove_function_parse_error(self):
        with patch("pathlib.Path.read_text", side_effect=SyntaxError("Invalid syntax")):
            result = remove_unused_function("test.py", "func", 1)
            assert result is False


class TestInteractiveSelection:
    @pytest.fixture
    def mock_console(self):
        return Mock()

    @pytest.fixture
    def sample_unused_items(self):
        """create fake sample unused items for testing"""
        functions = [
            {"name": "unused_func1", "file": "test1.py", "line": 10},
            {"name": "unused_func2", "file": "test2.py", "line": 20},
        ]
        imports = [
            {"name": "unused_import1", "file": "test1.py", "line": 1},
            {"name": "unused_import2", "file": "test2.py", "line": 2},
        ]
        return functions, imports

    def test_interactive_selection_unavailable(self, mock_console, sample_unused_items):
        functions, imports = sample_unused_items

        with patch("skylos.cli.INTERACTIVE_AVAILABLE", False):
            selected_functions, selected_imports = interactive_selection(
                mock_console, functions, imports
            )

        assert selected_functions == []
        assert selected_imports == []
        mock_console.print.assert_called_once()

    @patch("skylos.cli.inquirer")
    def test_interactive_selection_with_selections(
        self, mock_inquirer, mock_console, sample_unused_items
    ):
        functions, imports = sample_unused_items

        mock_inquirer.prompt.side_effect = [
            {"functions": [functions[0]]},
            {"imports": [imports[1]]},
        ]

        with patch("skylos.cli.INTERACTIVE_AVAILABLE", True):
            selected_functions, selected_imports = interactive_selection(
                mock_console, functions, imports
            )

    @patch("skylos.cli.inquirer")
    def test_interactive_selection_no_selections(
        self, mock_inquirer, mock_console, sample_unused_items
    ):
        functions, imports = sample_unused_items

        mock_inquirer.prompt.return_value = None

        with patch("skylos.cli.INTERACTIVE_AVAILABLE", True):
            selected_functions, selected_imports = interactive_selection(
                mock_console, functions, imports
            )

        assert selected_functions == []
        assert selected_imports == []

    def test_interactive_selection_empty_lists(self, mock_console):
        selected_functions, selected_imports = interactive_selection(
            mock_console, [], []
        )

        assert selected_functions == []
        assert selected_imports == []


class TestPrintBadge:
    @pytest.fixture
    def mock_logger(self):
        logger = Mock()
        logger.console = Mock()
        return logger

    def test_print_badge_zero_dead_code(self, mock_logger):
        """Test badge printing with zero dead code."""
        print_badge(0, mock_logger)

        calls = [c.args[0] for c in mock_logger.console.print.call_args_list]
        badge_call = next(
            (c for c in calls if isinstance(c, str) and "Dead_Code-Free" in c),
            None,
        )
        assert badge_call is not None
        assert "brightgreen" in badge_call

    def test_print_badge_with_dead_code(self, mock_logger):
        print_badge(5, mock_logger)

        calls = [c.args[0] for c in mock_logger.console.print.call_args_list]
        badge_call = next(
            (c for c in calls if isinstance(c, str) and "Dead_Code-5" in c),
            None,
        )
        assert badge_call is not None
        assert "orange" in badge_call


class TestMainFunction:
    @pytest.fixture
    def mock_skylos_result(self):
        return {
            "unused_functions": [
                {"name": "unused_func", "file": "test.py", "line": 10}
            ],
            "unused_imports": [{"name": "unused_import", "file": "test.py", "line": 1}],
            "unused_parameters": [],
            "unused_variables": [],
            "analysis_summary": {"total_files": 2, "excluded_folders": []},
        }

    def test_main_json_output(self, mock_skylos_result):
        """testing main function with JSON output"""
        test_args = ["cli.py", "test_path", "--json"]

        with (
            patch("sys.argv", test_args),
            patch("skylos.cli.run_analyze") as mock_analyze,
            patch("builtins.print") as mock_print,
            patch("skylos.cli.setup_logger"),
            patch("skylos.cli.Progress") as mock_progress,
        ):
            mock_progress.return_value.__enter__.return_value = Mock(add_task=Mock())
            mock_analyze.return_value = json.dumps(mock_skylos_result)

            main()

            mock_analyze.assert_called_once()
            mock_print.assert_called_once_with(json.dumps(mock_skylos_result))

    def test_main_verbose_output(self, mock_skylos_result):
        """with verbose"""
        test_args = ["cli.py", "test_path", "--verbose"]

        with (
            patch("sys.argv", test_args),
            patch("skylos.cli.run_analyze") as mock_analyze,
            patch("skylos.cli.setup_logger") as mock_setup_logger,
            patch("skylos.cli.Progress") as mock_progress,
        ):
            mock_logger = Mock()
            mock_logger.console = Mock()
            mock_setup_logger.return_value = mock_logger
            mock_analyze.return_value = json.dumps(mock_skylos_result)
            mock_progress.return_value.__enter__.return_value = Mock(add_task=Mock())

            main()

            mock_logger.setLevel.assert_called_with(logging.DEBUG)

    def test_main_analysis_error(self):
        test_args = ["cli.py", "test_path"]

        with (
            patch("sys.argv", test_args),
            patch("skylos.cli.run_analyze", side_effect=Exception("Analysis failed")),
            patch("skylos.cli.setup_logger") as mock_setup_logger,
            patch("skylos.cli.parse_exclude_folders", return_value=set()),
            patch("skylos.cli.Progress") as mock_progress,
        ):
            mock_logger = Mock()
            mock_logger.console = Mock()
            mock_setup_logger.return_value = mock_logger
            mock_progress.return_value.__enter__.return_value = Mock(add_task=Mock())

            with pytest.raises(SystemExit):
                main()

            mock_logger.error.assert_called_with(
                "Error during analysis: Analysis failed"
            )


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



from __future__ import annotations
import ast
import sys
from pathlib import Path
from .danger_sql.sql_flow import scan as scan_sql
from .danger_cmd.cmd_flow import scan as scan_cmd
from .danger_sql.sql_raw_flow import scan as scan_sql_raw
from .danger_net.ssrf_flow import scan as scan_ssrf
from .danger_fs.path_flow import scan as scan_path
from .danger_web.xss_flow import scan as scan_xss

ALLOWED_SUFFIXES = (".py", ".pyi", ".pyw")

DANGEROUS_CALLS = {
    "eval": ("SKY-D201", "HIGH", "Use of eval()"),
    "exec": ("SKY-D202", "HIGH", "Use of exec()"),
    "pickle.load": (
        "SKY-D203",
        "CRITICAL",
        "Untrusted deserialization via pickle.load",
    ),
    "pickle.loads": (
        "SKY-D204",
        "CRITICAL",
        "Untrusted deserialization via pickle.loads",
    ),
    "yaml.load": ("SKY-D205", "HIGH", "yaml.load without SafeLoader"),
    "hashlib.md5": ("SKY-D206", "MEDIUM", "Weak hash (MD5)"),
    "hashlib.sha1": ("SKY-D207", "MEDIUM", "Weak hash (SHA1)"),
    "requests.*": (
        "SKY-D208",
        "HIGH",
        "requests call with verify=False",
        {"kw_equals": {"verify": False}},
    ),
}


def _matches_rule(name, rule_key):
    if not name:
        return False
    if rule_key.endswith(".*"):
        return name.startswith(rule_key[:-2] + ".")
    return name == rule_key


def _kw_equals(node: ast.Call, requirements):
    if not requirements:
        return True
    kw_map = {}
    for kw in node.keywords or []:
        if kw.arg:
            kw_map[kw.arg] = kw.value

    for key, expected in requirements.items():
        val = kw_map.get(key)
        if not isinstance(val, ast.Constant):
            return False
        if val.value != expected:
            return False
    return True


def qualified_name_from_call(node: ast.Call):
    func = node.func
    parts = []
    while isinstance(func, ast.Attribute):
        parts.append(func.attr)
        func = func.value
    if isinstance(func, ast.Name):
        parts.append(func.id)
        parts.reverse()
        return ".".join(parts)
    return None


def _yaml_load_without_safeloader(node: ast.Call):
    name = qualified_name_from_call(node)
    if name != "yaml.load":
        return False

    for kw in node.keywords or []:
        if kw.arg == "Loader":
            text = ast.unparse(kw.value)
            return "SafeLoader" not in text
    return True


def _add_finding(findings, file_path: Path, node: ast.AST, rule_id, severity, message):
    findings.append(
        {
            "rule_id": rule_id,
            "severity": severity,
            "message": message,
            "file": str(file_path),
            "line": getattr(node, "lineno", 1),
            "col": getattr(node, "col_offset", 0),
        }
    )


def _scan_file(file_path: Path, findings):
    src = file_path.read_text(encoding="utf-8", errors="ignore")
    tree = ast.parse(src)

    scan_sql(tree, file_path, findings)
    scan_cmd(tree, file_path, findings)
    scan_sql_raw(tree, file_path, findings)
    scan_ssrf(tree, file_path, findings)
    scan_path(tree, file_path, findings)
    scan_xss(tree, file_path, findings)

    for node in ast.walk(tree):
        if not isinstance(node, ast.Call):
            continue

        name = qualified_name_from_call(node)
        if not name:
            continue

        for rule_key, tup in DANGEROUS_CALLS.items():
            rule_id = tup[0]
            severity = tup[1]
            message = tup[2]

            if len(tup) > 3:
                opts = tup[3]
            else:
                opts = None

            if not _matches_rule(name, rule_key):
                continue

            if rule_key == "yaml.load":
                if not _yaml_load_without_safeloader(node):
                    continue
            if opts and "kw_equals" in opts:
                if not _kw_equals(node, opts["kw_equals"]):
                    continue

            _add_finding(findings, file_path, node, rule_id, severity, message)
            break


def scan_ctx(_, files):
    findings = []

    for file_path in files:
        if file_path.suffix.lower() not in ALLOWED_SUFFIXES:
            continue

        try:
            _scan_file(file_path, findings)
        except Exception as e:
            print(f"Scan failed for {file_path}: {e}", file=sys.stderr)

    return findings



from tree_sitter import Language, QueryCursor, Query
import tree_sitter_typescript as tsts

try:
    TS_LANG = Language(tsts.language_typescript())
except Exception:
    TS_LANG = None

COMPLEXITY_NODES = {
    "if_statement",
    "for_statement",
    "while_statement",
    "switch_case",
    "catch_clause",
    "ternary_expression",
}


def scan_quality(root_node, source, file_path, threshold=10):
    findings = []
    if not TS_LANG:
        return []

    func_nodes = []
    query_str = """
    (function_declaration) @func
    (arrow_function) @func
    (method_definition) @func
    """

    try:
        query = Query(TS_LANG, query_str)
        cursor = QueryCursor(query)
        captures = cursor.captures(root_node)
        func_nodes = captures.get("func", [])
    except Exception:
        return []

    for func_node in func_nodes:
        complexity = _calc_complexity(func_node)

        if complexity > threshold:
            line = func_node.start_point[0] + 1
            name = "anonymous"
            try:
                name_node = func_node.child_by_field_name("name")
                if name_node:
                    name = source[name_node.start_byte : name_node.end_byte].decode(
                        "utf-8", errors="replace"
                    )
            except Exception:
                pass

            findings.append(
                {
                    "rule_id": "SKY-Q501",
                    "severity": "MEDIUM",
                    "message": f"Function '{name}' is too complex ({complexity})",
                    "file": str(file_path),
                    "line": line,
                    "col": 0,
                }
            )

    return findings


def _calc_complexity(node):
    count = 1
    cursor = node.walk()
    visited_children = False

    while True:
        if visited_children:
            if cursor.node.id == node.id:
                break
            if cursor.goto_next_sibling():
                visited_children = False
            elif cursor.goto_parent():
                visited_children = True
            else:
                break
        else:
            if cursor.node.type in COMPLEXITY_NODES:
                count += 1
            if cursor.goto_first_child():
                visited_children = False
            else:
                visited_children = True
    return count



import re
from pathlib import Path


class ImplicitRefTracker:
    
    def __init__(self):
        self.known_refs = set()
        self.pattern_refs = []
        self.f_string_patterns = {}
        self.coverage_hits = set()
    
    def should_mark_as_used(self, definition):
        simple_name = definition.simple_name
        
        if simple_name in self.known_refs:
            return True, 95, "dynamic reference"
        
        for pattern, confidence in self.pattern_refs:
            regex = "^" + pattern.replace("*", ".*") + "$"
            if re.match(regex, simple_name):
                return True, confidence, f"pattern '{pattern}'"
        
        if self.coverage_hits:
            def_file = Path(definition.filename).name
            for cov_file, line in self.coverage_hits:
                if cov_file.endswith(def_file) or def_file in cov_file:
                    if definition.line == line:
                        return True, 100, "executed (coverage)"
        
        return False, 0, None
    
    def load_coverage(self, coverage_file=".coverage"):
        path = Path(coverage_file)
        if not path.exists():
            return
        
        try:
            import sqlite3
            conn = sqlite3.connect(str(path))
            cursor = conn.cursor()
            
            cursor.execute("SELECT id, path FROM file")
            files = {}
            for row in cursor.fetchall():
                key = row[0]
                value = row[1]
                files[key] = value
            
            cursor.execute("SELECT file_id, numbits FROM line_bits")
            for file_id, numbits in cursor.fetchall():
                if file_id in files:
                    filename = files[file_id]
                    for byte_idx, byte in enumerate(numbits):
                        for bit_idx in range(8):
                            if byte & (1 << bit_idx):
                                line = byte_idx * 8 + bit_idx
                                self.coverage_hits.add((filename, line))
            
            conn.close()
        except Exception as e:
            pass

pattern_tracker = ImplicitRefTracker()



from .base import BaseAdapter
from .openai_adapter import OpenAIAdapter
from .anthropic_adapter import AnthropicAdapter


def get_adapter(model, api_key=None):
    model_lower = model.lower()

    if "claude" in model_lower:
        return AnthropicAdapter(model, api_key)

    return OpenAIAdapter(model, api_key)



#!/usr/bin/env python3
import ast
from pathlib import Path
import re
from skylos.control_flow import evaluate_static_condition, extract_constant_string
from skylos.implicit_refs import pattern_tracker

PYTHON_BUILTINS = {
    "print",
    "len",
    "str",
    "int",
    "float",
    "list",
    "dict",
    "set",
    "tuple",
    "range",
    "open",
    "reversed",
    "super",
    "object",
    "type",
    "enumerate",
    "zip",
    "map",
    "filter",
    "sorted",
    "sum",
    "min",
    "next",
    "iter",
    "bytes",
    "bytearray",
    "format",
    "round",
    "abs",
    "complex",
    "hash",
    "id",
    "bool",
    "callable",
    "getattr",
    "max",
    "all",
    "any",
    "setattr",
    "hasattr",
    "isinstance",
    "globals",
    "locals",
    "vars",
    "dir",
    "property",
    "classmethod",
    "staticmethod",
}
DYNAMIC_PATTERNS = {"getattr", "globals", "eval", "exec"}

## "🥚" hi :)

class Definition:
    def __init__(self, name, t, filename, line):
        self.name = name
        self.type = t
        self.filename = filename
        self.line = line
        self.simple_name = name.split(".")[-1]
        self.confidence = 100
        self.references = 0
        self.is_exported = False
        self.in_init = "__init__.py" in str(filename)

    def to_dict(self):
        if self.type == "method" and "." in self.name:
            parts = self.name.split(".")
            if len(parts) >= 3:
                output_name = ".".join(parts[-2:])
            else:
                output_name = self.name
        else:
            output_name = self.simple_name

        return {
            "name": output_name,
            "full_name": self.name,
            "simple_name": self.simple_name,
            "type": self.type,
            "file": str(self.filename),
            "basename": Path(self.filename).name,
            "line": self.line,
            "confidence": self.confidence,
            "references": self.references,
        }


class Visitor(ast.NodeVisitor):
    def __init__(self, mod, file):
        self.mod = mod
        self.file = file
        self.defs = []
        self.refs = []
        self.cls = None
        self.alias = {}
        self.dyn = set()
        self.exports = set()
        self.current_function_scope = []
        self.current_function_params = []
        self.local_var_maps = []
        self.in_cst_class = 0
        self.local_type_maps = []
        self._dataclass_stack = []
        self.dataclass_fields = set()
        self.first_read_lineno = {}
        self.instance_attr_types = {}
        self.local_constants = []

        self.pattern_tracker = pattern_tracker

    def add_def(self, name, t, line):
        found = False
        for d in self.defs:
            if d.name == name:
                found = True
                break
        if not found:
            self.defs.append(Definition(name, t, self.file, line))

    def add_ref(self, name):
        self.refs.append((name, self.file))

    def qual(self, name):
        if name in self.alias:
            return self.alias[name]
        if name in PYTHON_BUILTINS:
            if self.mod:
                mod_candidate = f"{self.mod}.{name}"
            else:
                mod_candidate = name
            if any(d.name == mod_candidate for d in self.defs):
                return mod_candidate

        if self.mod:
            return f"{self.mod}.{name}"
        else:
            return name

    def visit_Global(self, node):
        if self.current_function_scope and self.local_var_maps:
            for name in node.names:
                self.local_var_maps[-1][name] = f"{self.mod}.{name}"
        return

    def visit_annotation(self, node):
        if node is not None:
            if isinstance(node, ast.Constant) and isinstance(node.value, str):
                self.visit_string_annotation(node.value)
            elif hasattr(node, "s") and isinstance(node.s, str):
                self.visit_string_annotation(node.s)
            else:
                self.visit(node)

    def visit_string_annotation(self, annotation_str):
        if not isinstance(annotation_str, str):
            return

        try:
            parsed = ast.parse(annotation_str, mode="eval")
            self.visit(parsed.body)
        except SyntaxError:
            for tok in re.findall(r"[A-Za-z_][A-Za-z0-9_]*", annotation_str):
                self.add_ref(tok)

    def visit_Import(self, node):
        for a in node.names:
            full = a.name

            if a.asname:
                alias_name = a.asname
                target = full
            else:
                head = full.split(".", 1)[0]
                alias_name = head
                target = head

            self.alias[alias_name] = target
            self.add_def(target, "import", node.lineno)

    def visit_ImportFrom(self, node):
        if node.module is None:
            return
        
        for a in node.names:
            if a.name == "*":
                continue

            base = node.module
            if node.level:
                parts = self.mod.split(".")
                base = ".".join(parts[: -node.level]) + (
                    f".{node.module}" if node.module else ""
                )

            full = f"{base}.{a.name}"
            if a.asname:
                self.alias[a.asname] = full
                self.add_def(full, "import", node.lineno)
            else:
                self.alias[a.name] = full
                self.add_def(full, "import", node.lineno)
    
    def visit_If(self, node):
        condition = evaluate_static_condition(node.test)
        
        self.visit(node.test)

        if condition is True:
            for statement in node.body:
                self.visit(statement)
        elif condition is False:
            for statement in node.orelse:
                self.visit(statement)
        else:
            for statement in node.body:
                self.visit(statement)
            for statement in node.orelse:
                self.visit(statement)

    def visit_arguments(self, args):
        for arg in args.args:
            self.visit_annotation(arg.annotation)
        for arg in args.posonlyargs:
            self.visit_annotation(arg.annotation)

        for arg in args.kwonlyargs:
            self.visit_annotation(arg.annotation)
        if args.vararg:
            self.visit_annotation(args.vararg.annotation)
        if args.kwarg:
            self.visit_annotation(args.kwarg.annotation)
        for default in args.defaults:
            self.visit(default)
        for default in args.kw_defaults:
            if default:
                self.visit(default)

    def _get_decorator_name(self, deco):
        if isinstance(deco, ast.Name):
            return deco.id
        elif isinstance(deco, ast.Attribute):
            parent = self._get_decorator_name(deco.value)
            if parent:
                return f"{parent}.{deco.attr}"
            return deco.attr
        elif isinstance(deco, ast.Call):
            return self._get_decorator_name(deco.func)
        return None

    def visit_FunctionDef(self, node):
        outer_scope_prefix = (
            ".".join(self.current_function_scope) + "."
            if self.current_function_scope
            else ""
        )

        if self.cls:
            name_parts = [self.mod, self.cls, outer_scope_prefix + node.name]
        else:
            name_parts = [self.mod, outer_scope_prefix + node.name]

        qualified_name = ".".join(filter(None, name_parts))

        if self.cls:
            def_type = "method"
        else:
            def_type = "function"
        self.add_def(qualified_name, def_type, node.lineno)

        FRAMEWORK_DECORATORS = {
            'route', '.get', '.post', '.put', '.delete', '.patch',
            'fixture', 'pytest', 'task', 'celery', 'register',
            'subscriber', 'listener', 'handler', 'receiver', 'command'
        }

        for deco in node.decorator_list:
            deco_name = self._get_decorator_name(deco)
            if deco_name:
                if deco_name in ("property", "cached_property", "functools.cached_property"):
                    self.add_ref(qualified_name)
                elif deco_name.endswith((".setter", ".deleter")):
                    self.add_ref(qualified_name)
                elif any(keyword in deco_name.lower() for keyword in FRAMEWORK_DECORATORS):
                    self.add_ref(qualified_name)

            elif deco_name and deco_name.endswith((".setter", ".deleter")):
                self.add_ref(qualified_name)

        self.current_function_scope.append(node.name)
        self.local_var_maps.append({})
        self.local_type_maps.append({})

        self.local_constants.append({})

        old_params = self.current_function_params
        self.current_function_params = []

        for d_node in node.decorator_list:
            self.visit(d_node)

        for arg in node.args.args:
            param_name = f"{qualified_name}.{arg.arg}"
            self.add_def(param_name, "parameter", node.lineno)
            self.current_function_params.append((arg.arg, param_name))

        self.visit_arguments(node.args)
        self.visit_annotation(node.returns)

        for stmt in node.body:
            self.visit(stmt)

        self.current_function_scope.pop()
        self.current_function_params = old_params
        self.local_var_maps.pop()
        self.local_type_maps.pop()
        self.local_constants.pop()

    visit_AsyncFunctionDef = visit_FunctionDef

    def visit_ClassDef(self, node):
        cname = f"{self.mod}.{node.name}"
        self.add_def(cname, "class", node.lineno)

        is_cst = False
        is_dc = False

        for base in node.bases:
            base_name = ""

            if isinstance(base, ast.Attribute):
                base_name = base.attr

            elif isinstance(base, ast.Name):
                base_name = base.id
            self.visit(base)

            if base_name in {"CSTTransformer", "CSTVisitor"}:
                is_cst = True

        for keyword in node.keywords:
            self.visit(keyword.value)

        for decorator in node.decorator_list:

            def _is_dc(dec):
                if isinstance(dec, ast.Call):
                    target = dec.func
                else:
                    target = dec

                if isinstance(target, ast.Name):
                    return target.id == "dataclass"

                if isinstance(target, ast.Attribute):
                    return target.attr == "dataclass"

                return False

            if _is_dc(decorator):
                is_dc = True
            self.visit(decorator)

        prev = self.cls
        if is_cst:
            self.in_cst_class += 1

        self.cls = node.name
        self._dataclass_stack.append(is_dc)
        for b in node.body:
            self.visit(b)

        self.cls = prev
        self._dataclass_stack.pop()

        if is_cst:
            self.in_cst_class -= 1

    def visit_AnnAssign(self, node):
        self.visit_annotation(node.annotation)
        if node.value:
            self.visit(node.value)

        def _define(t):
            if isinstance(t, ast.Name):
                name_simple = t.id
                scope_parts = [self.mod]
                if self.cls:
                    scope_parts.append(self.cls)

                if self.current_function_scope:
                    scope_parts.extend(self.current_function_scope)
                prefix = ".".join(filter(None, scope_parts))
                if prefix:
                    var_name = f"{prefix}.{name_simple}"
                else:
                    var_name = name_simple

                self.add_def(var_name, "variable", t.lineno)
                if (
                    self._dataclass_stack
                    and self._dataclass_stack[-1]
                    and self.cls
                    and not self.current_function_scope
                ):
                    self.dataclass_fields.add(var_name)

                if self.current_function_scope and self.local_var_maps:
                    self.local_var_maps[-1][name_simple] = var_name

            elif isinstance(t, (ast.Tuple, ast.List)):
                for elt in t.elts:
                    _define(elt)

        _define(node.target)

    def visit_AugAssign(self, node):
        if isinstance(node.target, ast.Name):
            nm = node.target.id
            if (
                self.current_function_scope
                and self.local_var_maps
                and self.local_var_maps
                and nm in self.local_var_maps[-1]
            ):
                # self.add_ref(self.local_var_maps[-1][nm])
                fq = self.local_var_maps[-1][nm]
                self.add_ref(fq)
                var_name = fq

            else:
                self.add_ref(self.qual(nm))
                scope_parts = [self.mod]
                if self.cls:
                    scope_parts.append(self.cls)

                if self.current_function_scope:
                    scope_parts.extend(self.current_function_scope)
                prefix = ".".join(filter(None, scope_parts))
                if prefix:
                    var_name = f"{prefix}.{nm}"
                else:
                    var_name = nm

            self.add_def(var_name, "variable", node.lineno)
            if self.current_function_scope and self.local_var_maps:
                self.local_var_maps[-1][nm] = var_name
        else:
            self.visit(node.target)
        self.visit(node.value)

    def visit_Subscript(self, node):
        self.visit(node.value)
        self.visit(node.slice)

    def visit_Slice(self, node):
        if node.lower:
            self.visit(node.lower)
        if node.upper:
            self.visit(node.upper)
        if node.step:
            self.visit(node.step)

    def _should_skip_variable_def(self, name_simple):
        if (
            name_simple == "METADATA_DEPENDENCIES"
            and self.cls
            and self.in_cst_class > 0
        ):
            return True

        if (
            name_simple == "__all__"
            and not self.current_function_scope
            and not self.cls
        ):
            return True

        return False

    def _compute_variable_name(self, name_simple):
        scope_parts = [self.mod]
        if self.cls:
            scope_parts.append(self.cls)
        if self.current_function_scope:
            scope_parts.extend(self.current_function_scope)

        if (
            self.current_function_scope
            and self.local_var_maps
            and name_simple in self.local_var_maps[-1]
        ):
            return self.local_var_maps[-1][name_simple]

        prefix = ".".join(filter(None, scope_parts))
        if prefix:
            return f"{prefix}.{name_simple}"
        return name_simple

    def _process_target_for_def(self, target_node):
        if isinstance(target_node, ast.Name):
            name_simple = target_node.id

            if self._should_skip_variable_def(name_simple):
                return

            var_name = self._compute_variable_name(name_simple)
            self.add_def(var_name, "variable", target_node.lineno)

            if (
                self.current_function_scope
                and self.local_var_maps
                and name_simple not in self.local_var_maps[-1]
            ):
                self.local_var_maps[-1][name_simple] = var_name

        elif isinstance(target_node, (ast.Tuple, ast.List)):
            for elt in target_node.elts:
                self._process_target_for_def(elt)

    def _extract_string_value(self, elt):
        if isinstance(elt, ast.Constant) and isinstance(elt.value, str):
            return elt.value
        if hasattr(elt, "s") and isinstance(elt.s, str):
            return elt.s
        return None

    def _process_dunder_all_exports(self, node):
        for target in node.targets:
            if not isinstance(target, ast.Name):
                continue
            if target.id != "__all__":
                continue
            if not isinstance(node.value, (ast.List, ast.Tuple)):
                continue

            for elt in node.value.elts:
                value = self._extract_string_value(elt)
                if value is None:
                    continue

                if self.mod:
                    export_name = f"{self.mod}.{value}"
                else:
                    export_name = value

                self.add_ref(export_name)
                self.add_ref(value)

    def _resolve_callee_fqname_from_name(self, callee):
        return self.alias.get(callee.id, self.qual(callee.id))

    def _resolve_callee_fqname_from_attribute(self, callee):
        parts = []
        cur = callee

        while isinstance(cur, ast.Attribute):
            parts.append(cur.attr)
            cur = cur.value

        if not isinstance(cur, ast.Name):
            return None

        head = self.alias.get(cur.id, self.qual(cur.id))
        if not head:
            return None

        return ".".join([head] + list(reversed(parts)))

    def _resolve_callee_fqname(self, callee):
        if isinstance(callee, ast.Name):
            return self._resolve_callee_fqname_from_name(callee)

        if isinstance(callee, ast.Attribute):
            return self._resolve_callee_fqname_from_attribute(callee)

        return None

    def _mark_target_type(self, target, fqname):
        if isinstance(target, ast.Name):
            self.local_type_maps[-1][target.id] = fqname
        elif isinstance(target, (ast.Tuple, ast.List)):
            for elt in target.elts:
                self._mark_target_type(elt, fqname)

    def _try_infer_types_from_call(self, node):
        if not isinstance(node.value, ast.Call):
            return

        if not self.current_function_scope:
            return
        if not self.local_type_maps:
            return

        call_node = node.value
        if not hasattr(call_node, "func"):
            return

        callee = call_node.func
        fqname = self._resolve_callee_fqname(callee)

        if not fqname:
            return

        for target in node.targets:
            self._mark_target_type(target, fqname)

    def _extract_class_name_from_call(self, call_func):
        if isinstance(call_func, ast.Name):
            return call_func.id
        if isinstance(call_func, ast.Attribute):
            return call_func.attr
        return None

    def _track_instance_attr_types(self, node):
        if not self.cls:
            return

        if not isinstance(node.value, ast.Call):
            return

        for target in node.targets:
            if not isinstance(target, ast.Attribute):
                continue
            if not isinstance(target.value, ast.Name):
                continue
            if target.value.id != "self":
                continue

            call_func = node.value.func
            class_name = self._extract_class_name_from_call(call_func)

            if not class_name:
                continue
            if not class_name[0].isupper():
                continue

            owner = f"{self.mod}.{self.cls}" if self.mod else self.cls
            attr_key = f"{owner}.{target.attr}"
            qualified_class = self.alias.get(class_name, self.qual(class_name))
            self.instance_attr_types[attr_key] = qualified_class

    def visit_Assign(self, node):
        
        const_val = extract_constant_string(node.value)
        if const_val is not None:
            if len(self.local_constants) > 0:
                for t in node.targets:
                    if isinstance(t, ast.Name):
                        self.local_constants[-1][t.id] = const_val
        
        if isinstance(node.value, ast.JoinedStr):
            pattern = self._extract_fstring_pattern(node.value)
            if pattern:
                for t in node.targets:
                    if isinstance(t, ast.Name):
                        self.pattern_tracker.f_string_patterns[t.id] = pattern

        for target in node.targets:
            self._process_target_for_def(target)

        self._process_dunder_all_exports(node)
        self._try_infer_types_from_call(node)
        self.generic_visit(node)
        self._track_instance_attr_types(node)

    def _extract_fstring_pattern(self, node: ast.JoinedStr):
        """Convert f'handle_{x}' to 'handle_*'"""
        parts = []
        has_var = False
        for value in node.values:
            if isinstance(value, ast.Constant):
                parts.append(str(value.value))
            elif isinstance(value, ast.FormattedValue):
                parts.append("*")
                has_var = True
        return "".join(parts) if has_var else None
    
    def visit_Call(self, node):
        self.generic_visit(node)

        if (
            isinstance(node.func, ast.Name)
            and node.func.id in ("getattr", "hasattr")
            and len(node.args) >= 2
        ):
            
            attr_name = None

            if isinstance(node.args[1], ast.Name) and self.local_constants:
                attr_name = self.local_constants[-1].get(node.args[1].id)

            if not attr_name:
                attr_name = extract_constant_string(node.args[1])
            
            if attr_name:
                self.add_ref(attr_name)

                if isinstance(node.args[0], ast.Name):
                    module_name = node.args[0].id
                    if module_name != "self":
                        qualified_name = f"{self.qual(module_name)}.{attr_name}"
                        self.add_ref(qualified_name)
            
            # elif isinstance(node.args[0], ast.Name):
            #     target_name = node.args[0].id
            #     if target_name != "self":
            #         if self.mod:
            #             self.dyn.add(self.mod.split(".")[0])
            #         else:
            #             self.dyn.add("")

            elif isinstance(node.args[0], ast.Name):
                target_name = node.args[0].id
                if target_name != "self":
                    if isinstance(node.args[1], ast.Name):
                        var_name = node.args[1].id
                        if var_name in self.pattern_tracker.f_string_patterns:
                            pattern = self.pattern_tracker.f_string_patterns[var_name]
                            self.pattern_tracker.pattern_refs.append((pattern, 70))
                        elif self.local_constants and var_name in self.local_constants[-1]:
                            val = self.local_constants[-1][var_name]
                            self.pattern_tracker.known_refs.add(val)

        elif isinstance(node.func, ast.Name) and node.func.id == "globals":
            parent = getattr(node, "parent", None)
            if (
                isinstance(parent, ast.Subscript)
                and isinstance(parent.slice, ast.Constant)
                and isinstance(parent.slice.value, str)
            ):
                func_name = parent.slice.value
                self.add_ref(func_name)
                self.add_ref(f"{self.mod}.{func_name}")

        elif isinstance(node.func, ast.Name) and node.func.id in ("eval", "exec"):
            root_mod = ""
            if self.mod:
                root_mod = self.mod.split(".")[0]
            self.dyn.add(root_mod)

        elif (
            isinstance(node.func, ast.Attribute)
            and isinstance(node.func.value, ast.Call)
            and isinstance(node.func.value.func, ast.Name)
            and node.func.value.func.id == "super"
        ):
            method_name = node.func.attr
            if self.cls:
                if self.mod:
                    owner = f"{self.mod}.{self.cls}"
                else:
                    owner = self.cls
                self.add_ref(f"{owner}.{method_name}")

    def visit_Name(self, node):
        if isinstance(node.ctx, ast.Load):
            for param_name, param_full_name in self.current_function_params:
                if node.id == param_name:
                    self.first_read_lineno.setdefault(param_full_name, node.lineno)
                    self.add_ref(param_full_name)
                    return

            if (
                self.current_function_scope
                and self.local_var_maps
                and node.id in self.local_var_maps[-1]
            ):
                fq = self.local_var_maps[-1][node.id]
                self.first_read_lineno.setdefault(fq, node.lineno)
                self.add_ref(fq)
                return

            qualified = self.qual(node.id)
            self.first_read_lineno.setdefault(qualified, node.lineno)
            self.add_ref(qualified)
            if node.id in DYNAMIC_PATTERNS:
                self.dyn.add(self.mod.split(".")[0])

    def visit_Attribute(self, node):
        self.generic_visit(node)

        if not isinstance(node.ctx, ast.Load):
            return

        if isinstance(node.value, ast.Name):
            base = node.value.id

            param_hit = None
            for param_name, param_full in self.current_function_params:
                if base == param_name:
                    param_hit = (param_name, param_full)
                    break

            if param_hit:
                self.add_ref(param_hit[1])

            if self.cls and base in {"self", "cls"}:
                if self.mod:
                    owner = f"{self.mod}.{self.cls}"
                else:
                    owner = self.cls

                self.add_ref(f"{owner}.{node.attr}")
                return

            if (
                self.current_function_scope
                and self.local_type_maps
                and self.local_type_maps[-1].get(base)
            ):
                self.add_ref(f"{self.local_type_maps[-1][base]}.{node.attr}")
                return

            self.add_ref(f"{self.qual(base)}.{node.attr}")

        elif isinstance(node.value, ast.Call):
            call_func = node.value.func
            class_name = None
            qualified_class = None

            if isinstance(call_func, ast.Name):
                class_name = call_func.id
                if class_name and class_name[0].isupper():
                    qualified_class = self.alias.get(class_name, self.qual(class_name))

            elif isinstance(call_func, ast.Attribute):
                class_name = call_func.attr
                if class_name and class_name[0].isupper():
                    if isinstance(call_func.value, ast.Name):
                        base = call_func.value.id
                        base_resolved = self.alias.get(base, base)
                        qualified_class = f"{base_resolved}.{class_name}"
                    else:
                        qualified_class = self.alias.get(
                            class_name, self.qual(class_name)
                        )

            if qualified_class:
                self.add_ref(f"{qualified_class}.{node.attr}")

        elif isinstance(node.value, ast.Attribute):
            inner = node.value
            if (
                isinstance(inner.value, ast.Name)
                and inner.value.id == "self"
                and self.cls
            ):
                owner = f"{self.mod}.{self.cls}" if self.mod else self.cls
                attr_key = f"{owner}.{inner.attr}"
                if attr_key in self.instance_attr_types:
                    type_name = self.instance_attr_types[attr_key]
                    self.add_ref(f"{type_name}.{node.attr}")

    def visit_NamedExpr(self, node):
        self.visit(node.value)
        if isinstance(node.target, ast.Name):
            nm = node.target.id
            scope_parts = [self.mod]

            if self.cls:
                scope_parts.append(self.cls)

            if self.current_function_scope:
                scope_parts.extend(self.current_function_scope)

            prefix = ".".join(filter(None, scope_parts))
            var_name = f"{prefix}.{nm}" if prefix else nm

            self.add_def(var_name, "variable", node.lineno)
            if self.current_function_scope and self.local_var_maps:
                self.local_var_maps[-1][nm] = var_name
            self.add_ref(var_name)

    def visit_keyword(self, node):
        self.visit(node.value)

    def visit_withitem(self, node):
        self.visit(node.context_expr)
        if node.optional_vars:
            self.visit(node.optional_vars)

    def visit_ExceptHandler(self, node):
        if node.type:
            self.visit(node.type)
        for stmt in node.body:
            self.visit(stmt)

    def generic_visit(self, node):
        for field, value in ast.iter_fields(node):
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, ast.AST):
                        item.parent = node
                        self.visit(item)
            elif isinstance(value, ast.AST):
                value.parent = node
                self.visit(value)



# tests/test_cli_coverage.py
import pytest
from unittest.mock import patch, MagicMock
import subprocess
import sys


class TestCoverageFlag:
    """Tests for the --coverage CLI flag."""
    
    @patch('skylos.cli.subprocess.run')
    @patch('skylos.cli.run_analyze')
    def test_coverage_runs_pytest_first(self, mock_analyze, mock_run):
        """--coverage should run pytest with coverage before analysis."""
        from skylos.cli import main
        
        # Mock successful pytest run
        mock_run.return_value = MagicMock(returncode=0)
        mock_analyze.return_value = '{"unused_functions": [], "unused_imports": [], "unused_classes": [], "unused_variables": [], "unused_parameters": [], "analysis_summary": {"total_files": 1}}'
        
        with patch.object(sys, 'argv', ['skylos', '.', '--coverage', '--json']):
            try:
                main()
            except SystemExit:
                pass
        
        # Verify coverage run was called
        calls = mock_run.call_args_list
        assert any(
            'coverage' in str(call) and 'pytest' in str(call)
            for call in calls
        )
    
    @patch('skylos.cli.subprocess.run')
    @patch('skylos.cli.run_analyze')
    def test_coverage_falls_back_to_unittest(self, mock_analyze, mock_run):
        """If pytest fails, should fall back to unittest."""
        from skylos.cli import main
        
        # First call (pytest) fails, second call (unittest) succeeds
        mock_run.side_effect = [
            MagicMock(returncode=1),  # pytest fails
            MagicMock(returncode=0),  # unittest succeeds
        ]
        mock_analyze.return_value = '{"unused_functions": [], "unused_imports": [], "unused_classes": [], "unused_variables": [], "unused_parameters": [], "analysis_summary": {"total_files": 1}}'
        
        with patch.object(sys, 'argv', ['skylos', '.', '--coverage', '--json']):
            try:
                main()
            except SystemExit:
                pass
        
        # Should have called both pytest and unittest
        assert mock_run.call_count == 2
        
        calls = [str(c) for c in mock_run.call_args_list]
        assert any('pytest' in c for c in calls)
        assert any('unittest' in c for c in calls)
    
    @patch('skylos.cli.subprocess.run')
    @patch('skylos.cli.run_analyze')
    def test_coverage_runs_before_analysis(self, mock_analyze, mock_run):
        """Coverage collection must happen BEFORE analysis."""
        from skylos.cli import main
        
        call_order = []
        
        def track_subprocess(*args, **kwargs):
            call_order.append('coverage')
            return MagicMock(returncode=0)
        
        def track_analyze(*args, **kwargs):
            call_order.append('analyze')
            return '{"unused_functions": [], "unused_imports": [], "unused_classes": [], "unused_variables": [], "unused_parameters": [], "analysis_summary": {"total_files": 1}}'
        
        mock_run.side_effect = track_subprocess
        mock_analyze.side_effect = track_analyze
        
        with patch.object(sys, 'argv', ['skylos', '.', '--coverage', '--json']):
            try:
                main()
            except SystemExit:
                pass
        
        # Coverage must come before analyze
        assert call_order.index('coverage') < call_order.index('analyze')
    
    @patch('skylos.cli.subprocess.run')
    @patch('skylos.cli.run_analyze')
    def test_coverage_uses_project_root_cwd(self, mock_analyze, mock_run):
        """Coverage should run in the project root directory."""
        from skylos.cli import main
        
        mock_run.return_value = MagicMock(returncode=0)
        mock_analyze.return_value = '{"unused_functions": [], "unused_imports": [], "unused_classes": [], "unused_variables": [], "unused_parameters": [], "analysis_summary": {"total_files": 1}}'
        
        with patch.object(sys, 'argv', ['skylos', '/some/project', '--coverage', '--json']):
            try:
                main()
            except SystemExit:
                pass
        
        # Check that cwd was passed to subprocess.run
        call_kwargs = mock_run.call_args_list[0][1]
        assert 'cwd' in call_kwargs
    
    @patch('skylos.cli.run_analyze')
    def test_no_coverage_flag_skips_coverage(self, mock_analyze):
        """Without --coverage flag, coverage should not run."""
        from skylos.cli import main
        
        mock_analyze.return_value = '{"unused_functions": [], "unused_imports": [], "unused_classes": [], "unused_variables": [], "unused_parameters": [], "analysis_summary": {"total_files": 1}}'
        
        with patch('skylos.cli.subprocess.run') as mock_run:
            with patch.object(sys, 'argv', ['skylos', '.', '--json']):
                try:
                    main()
                except SystemExit:
                    pass
            
            # subprocess.run should not be called for coverage
            coverage_calls = [
                c for c in mock_run.call_args_list
                if 'coverage' in str(c)
            ]
            assert len(coverage_calls) == 0
    
    @patch('skylos.cli.subprocess.run')
    @patch('skylos.cli.run_analyze')
    def test_coverage_with_other_flags(self, mock_analyze, mock_run):
        """--coverage should work with other flags like --danger."""
        from skylos.cli import main
        
        mock_run.return_value = MagicMock(returncode=0)
        mock_analyze.return_value = '{"unused_functions": [], "unused_imports": [], "unused_classes": [], "unused_variables": [], "unused_parameters": [], "danger": [], "analysis_summary": {"total_files": 1}}'
        
        with patch.object(sys, 'argv', ['skylos', '.', '--coverage', '--danger', '--json']):
            try:
                main()
            except SystemExit:
                pass
        
        # Coverage should still run
        assert mock_run.called
        
        # Analysis should include danger flag
        mock_analyze.assert_called_once()
        call_kwargs = mock_analyze.call_args[1]
        assert call_kwargs.get('enable_danger') is True


class TestCoverageIntegration:
    """Integration tests for coverage + analysis flow."""
    
    @patch('skylos.implicit_refs.pattern_tracker')
    @patch('skylos.cli.subprocess.run')
    @patch('skylos.cli.run_analyze')
    def test_coverage_data_loaded_before_analysis(self, mock_analyze, mock_run, mock_tracker):
        """After coverage runs, the .coverage file should be loaded."""
        from skylos.cli import main
        from pathlib import Path
        
        mock_run.return_value = MagicMock(returncode=0)
        mock_analyze.return_value = '{"unused_functions": [], "unused_imports": [], "unused_classes": [], "unused_variables": [], "unused_parameters": [], "analysis_summary": {"total_files": 1}}'
        
        # Simulate .coverage file existing after coverage run
        with patch.object(Path, 'exists', return_value=True):
            with patch.object(sys, 'argv', ['skylos', '.', '--coverage', '--json']):
                try:
                    main()
                except SystemExit:
                    pass
        
        # Analysis should have been called (which loads coverage internally)
        assert mock_analyze.called


import unittest
from pathlib import Path
from skylos.analyzer import Skylos


class TestPathExclusion(unittest.TestCase):
    def setUp(self):
        self.analyzer = Skylos()
        self.root = Path(".")

    def test_exclude_nested_folder_regression(self):
        excludes = ["src/legacy"]

        file_path = Path("src/legacy/old_main.py")
        self.assertTrue(
            self.analyzer._should_exclude_file(file_path, self.root, excludes),
            "Failed to exclude file inside nested folder 'src/legacy'",
        )

        deep_file = Path("src/legacy/utils/helpers.py")
        self.assertTrue(
            self.analyzer._should_exclude_file(deep_file, self.root, excludes),
            "Failed to exclude file deeply nested inside 'src/legacy'",
        )

    def test_exclude_nested_folder_false_positives(self):
        excludes = ["src/legacy"]

        similar_path = Path("src/legacy_v2/new.py")
        self.assertFalse(
            self.analyzer._should_exclude_file(similar_path, self.root, excludes),
            "Incorrectly excluded a folder sharing a prefix ('src/legacy_v2')",
        )

        diff_path = Path("src/modern/main.py")
        self.assertFalse(
            self.analyzer._should_exclude_file(diff_path, self.root, excludes),
            "Incorrectly excluded a non-matching sibling folder",
        )

    def test_windows_path_normalization(self):
        excludes = ["src/legacy"]

        file_path = Path("src") / "legacy" / "windows.py"

        self.assertTrue(
            self.analyzer._should_exclude_file(file_path, self.root, excludes),
            "Failed to exclude path constructed with OS separators",
        )

    def test_existing_wildcard_logic(self):
        excludes = ["*pycache__"]
        file_path = Path("src/__pycache__/cache.py")

        self.assertTrue(
            self.analyzer._should_exclude_file(file_path, self.root, excludes),
            "Regression: Wildcard exclusion failed",
        )

    def test_multiple_excludes_logic(self):
        excludes = ["src/legacy", "venv"]

        file_path = Path("venv/lib/site-packages/pkg.py")

        self.assertTrue(
            self.analyzer._should_exclude_file(file_path, self.root, excludes),
            "Regression: Loop stopped prematurely (did not check second exclude item)",
        )


if __name__ == "__main__":
    unittest.main()



import os
import re
from skylos.adapters import get_adapter


class Fixer:
    def __init__(self, api_key=None, model="gpt-4.1"):
        self.api_key = api_key or os.getenv("OPENAI_API_KEY")
        self.model = model

        try:
            self.adapter = get_adapter(model, self.api_key)
        except Exception as e:
            self.adapter = None
            self.init_error = str(e)

    def _add_line_numbers(self, source_code):
        lines = source_code.splitlines()
        numbered_lines = []
        for i, line in enumerate(lines, start=1):
            # format eg: "  10 | import os"
            numbered_lines.append(f"{i:4d} | {line}")
        return "\n".join(numbered_lines)

    def _get_relevant_context(self, source_code, defs_map):
        used_names = set(re.findall(r"\b[a-zA-Z_][a-zA-Z0-9_]*\b", source_code))
        relevant_tools = []

        for name, d in defs_map.items():
            if isinstance(d, dict):
                dname = d.get("name")
                dtype = d.get("type")
            else:
                dname = getattr(d, "name", name)
                dtype = getattr(d, "type", None)

            if dname in used_names:
                relevant_tools.append(f"- {dname} ({dtype})")

        if not relevant_tools:
            return "No internal dependencies detected."
        return "\n".join(relevant_tools[:100])

    def generate_prompt(self, source_code, error_line, error_msg, defs_map):
        lines = source_code.splitlines()
        start = max(0, error_line - 10)
        end = min(len(lines), error_line + 10)
        snippet = "\n".join(
            [f"{i + 1}: {line}" for i, line in enumerate(lines) if start <= i <= end]
        )

        tools_str = self._get_relevant_context(source_code, defs_map)

        prompt = f"""
You are an expert Python developer. Fix this specific error.

ERROR: Line {error_line}: {error_msg}
CODE:
```python
{snippet}

CONTEXT (These things exist): {tools_str}

OUTPUT FORMAT: ---PROBLEM--- (Summary) ---CHANGE--- (Summary) ---CODE--- (Fixed code block only) 
"""
        return prompt

    def audit_file(self, source_code, defs_map):
        if not self.adapter:
            return f"Error: {getattr(self, 'init_error', 'Adapter not initialized')}"

        tools_str = self._get_relevant_context(source_code, defs_map)
        numbered_code = self._add_line_numbers(source_code)

        prompt = f"""
You are Skylos AI, an advanced code auditor. Your job is to catch issues that the Static Analyzer (AST) might have missed. You are an INDEPENDENT pair of eyes.
SOURCE CODE (Line numbers provided on left):
```text
{numbered_code}
```

VERIFIED REPO CONTEXT (These functions/classes definitely exist): {tools_str}

INSTRUCTIONS - HUNT FOR THESE SPECIFIC FAILURES. THIS IS A CHALLENGE:

1. DEAD CODE: Are there imports, functions, or variables that look unused?

2. SECRETS:HARDCODED SECRETS (CRITICAL): Are there any hardcoded secrets (API keys, passwords, tokens):

    -> LOOK INSIDE arguments: os.getenv("VAR", "password123") is a LEAK. Flag "password123".
    -> LOOK INSIDE variables: key = "sk-..." is a LEAK.
    -> LOOK INSIDE comments: # pass: admin is a LEAK.

3. DANGEROUS CODE: Security risks (eval, sql injection, hardcoded secrets, shell=True).

4. VIBE CODING / HALLUCINATIONS: Is it calling a function that is NOT in the 'Verified Repo Context' and NOT a standard Python library?

5. QUALITY: Is the logic confusing? Are there bare exceptions? Infinite loops?

OUTPUT FORMAT: If the code is perfect, output: " No issues found." If you find issues, list them strictly like this:

[DEAD CODE] Line X: Description [SECURITY] Line X: Description [SECRETS] Line X: Description [HALLUCINATION] Line X: Description [QUALITY] Line X: Description 
"""

        instructions = "You are a strict code auditor."
        return self.adapter.complete(instructions, prompt)

    def fix_bug(self, source_code, error_line, error_msg, defs_map):
        if not self.adapter:
            return {
                "error": f"Error: {getattr(self, 'init_error', 'Adapter not initialized')}"
            }

        prompt = self.generate_prompt(source_code, error_line, error_msg, defs_map)
        instructions = "You are a strict code repair agent. Output only code."

        try:
            raw_text = self.adapter.complete(instructions, prompt)
            problem = "Issue detected"
            change = "Applied fix"
            code = raw_text

            if (
                "---PROBLEM---" in raw_text
                and "---CHANGE---" in raw_text
                and "---CODE---" in raw_text
            ):
                parts = raw_text.split("---CHANGE---")
                problem_part = parts[0].replace("---PROBLEM---", "").strip()
                rest = parts[1]
                code_parts = rest.split("---CODE---")
                change = code_parts[0].strip()
                code = code_parts[1].strip()
                problem = problem_part

            code = code.replace("```python", "").replace("```", "").strip()
            return {"code": code, "problem": problem, "change": change}

        except Exception as e:
            return {"error": f"Error calling OpenAI: {str(e)}"}



class BaseAdapter:
    def __init__(self, model, api_key):
        self.model = model
        self.api_key = api_key

    def complete(self, _, user_prompt):
        raise NotImplementedError("Subclasses must implement this method")



import sys
import subprocess
from rich.console import Console
from rich.prompt import Confirm, Prompt

try:
    import inquirer

    INTERACTIVE = True
except ImportError:
    INTERACTIVE = False

console = Console()


def run_cmd(cmd_list, error_msg="Git command failed"):
    try:
        result = subprocess.run(cmd_list, check=True, capture_output=True, text=True)
        return result.stdout.strip()
    except subprocess.CalledProcessError as e:
        console.print(f"[bold red]Error:[/bold red] {error_msg}\n[dim]{e.stderr}[/dim]")
        return None


def get_git_status():
    out = run_cmd(
        ["git", "status", "--porcelain"], "Could not get git status. Is this a repo?"
    )
    if not out:
        return []

    files = []
    for line in out.splitlines():
        if len(line) > 3:
            files.append(line[3:])
    return files


def run_push():
    console.print("[dim]Pushing to remote...[/dim]")
    try:
        subprocess.run(["git", "push"], check=True)
        console.print("[bold green] Deployment Complete. Code is live.[/bold green]")
    except subprocess.CalledProcessError:
        console.print(
            "[bold red] Push failed. Check your git remote settings.[/bold red]"
        )


def start_deployment_wizard():
    if not INTERACTIVE:
        console.print(
            "[yellow]Install 'inquirer' (pip install inquirer) to use interactive deployment.[/yellow]"
        )
        return

    console.print("\n[bold cyan] Skylos Deployment Wizard[/bold cyan]")

    files = get_git_status()
    if not files:
        console.print("[green]Working tree is clean.[/green]")
        if Confirm.ask("Push existing commits?"):
            run_push()
        return

    q_scope = [
        inquirer.List(
            "scope",
            message="What do you want to stage?",
            choices=[
                "All changed files",
                "Select files manually",
                "Skip commit (Push only)",
            ],
        ),
    ]
    ans_scope = inquirer.prompt(q_scope)
    if not ans_scope:
        return

    if ans_scope["scope"] == "Select files manually":
        q_files = [inquirer.Checkbox("files", message="Select files", choices=files)]
        ans_files = inquirer.prompt(q_files)
        if not ans_files or not ans_files["files"]:
            console.print("[red]No files selected.[/red]")
            return
        run_cmd(["git", "add"] + ans_files["files"])
        console.print(f"[green]Staged {len(ans_files['files'])} files.[/green]")

    elif ans_scope["scope"] == "All changed files":
        run_cmd(["git", "add", "."])
        console.print("[green]Staged all files.[/green]")

    if ans_scope["scope"] != "Skip commit (Push only)":
        msg = Prompt.ask("[bold green]Enter commit message[/bold green]")
        if not msg:
            console.print("[red]Commit message required.[/red]")
            return
        if run_cmd(["git", "commit", "-m", msg]):
            console.print("[green]✓ Committed.[/green]")

    if Confirm.ask("Ready to git push?"):
        run_push()


def check_gate(results, config):
    gate_cfg = config.get("gate", {})

    danger = results.get("danger", [])
    secrets = results.get("secrets", [])
    quality = results.get("quality", [])

    reasons = []

    criticals = []
    for f in danger:
        if f.get("severity") == "CRITICAL":
            criticals.append(f)

    if gate_cfg.get("fail_on_critical") and (criticals or secrets):
        if criticals:
            reasons.append(f"Found {len(criticals)} CRITICAL security issues")
        if secrets:
            reasons.append(f"Found {len(secrets)} Secrets")

    total_sec = len(danger)
    limit_sec = gate_cfg.get("max_security", 0)
    if total_sec > limit_sec:
        reasons.append(f"Security issues ({total_sec}) exceed limit ({limit_sec})")

    total_qual = len(quality)
    limit_qual = gate_cfg.get("max_quality", 10)
    if total_qual > limit_qual:
        reasons.append(f"Quality issues ({total_qual}) exceed limit ({limit_qual})")

    return (len(reasons) == 0), reasons


def run_gate_interaction(results, config, command_to_run):
    passed, reasons = check_gate(results, config)

    if passed:
        console.print("\n[bold green] Skylos Gate Passed.[/bold green]")
        if command_to_run:
            console.print(f"[dim]Running: {' '.join(command_to_run)}[/dim]")
            subprocess.run(command_to_run)
        else:
            start_deployment_wizard()
        return 0

    console.print("\n[bold red] Skylos Gate Failed![/bold red]")
    for reason in reasons:
        console.print(f" - {reason}")

    if config.get("gate", {}).get("strict"):
        console.print("[bold red]Strict mode enabled. Cannot bypass.[/bold red]")
        return 1

    if sys.stdout.isatty():
        if Confirm.ask(
            "\n[bold yellow]Do you want to bypass checks and proceed anyway?[/bold yellow]"
        ):
            console.print("[yellow]⚠ Bypassing Gate...[/yellow]")
            if command_to_run:
                subprocess.run(command_to_run)
            else:
                start_deployment_wizard()
            return 0

    return 1



import re

PENALTIES = {
    "private_name": 80,
    "dunder_or_magic": 100,
    "underscored_var": 100,
    "in_init_file": 15,
    "dynamic_module": 40,
    "test_related": 100,
    "framework_magic": 40,
}

TEST_FILE_RE = re.compile(r"(?:^|[/\\])tests?[/\\]|_test\.py$", re.I)
TEST_IMPORT_RE = re.compile(r"^(pytest|unittest|nose|mock|responses)(\.|$)")
TEST_DECOR_RE = re.compile(
    r"""^(
    pytest\.(fixture|mark) |
    patch(\.|$) |
    responses\.activate |
    freeze_time
)$""",
    re.X,
)

AUTO_CALLED = {
    "__init__",
    "__init__",
    "__new__",
    "__del__",
    "__init_subclass__",
    "__set_name__",
    "__enter__",
    "__exit__",
    "__iter__",
    "__next__",
    "__len__",
    "__getitem__",
    "__setitem__",
    "__delitem__",
    "__contains__",
    "__missing__",
    "__getattr__",
    "__setattr__",
    "__delattr__",
    "__getattribute__",
    "__str__",
    "__repr__",
    "__format__",
    "__bytes__",
    "__hash__",
    "__bool__",
}

TEST_METHOD_PATTERN = re.compile(r"^test_\w+$")

UNITTEST_LIFECYCLE_METHODS = {
    "setUp",
    "tearDown",
    "setUpClass",
    "tearDownClass",
    "setUpModule",
    "tearDownModule",
}

FRAMEWORK_FILE_RE = re.compile(r"(?:views|handlers|endpoints|routes|api)\.py$", re.I)

DEFAULT_EXCLUDE_FOLDERS = {
    "__pycache__",
    ".git",
    ".pytest_cache",
    ".mypy_cache",
    ".tox",
    "htmlcov",
    ".coverage",
    "build",
    "dist",
    "*.egg-info",
    "venv",
    ".venv",
}


def is_test_path(p):
    return bool(TEST_FILE_RE.search(str(p)))


def is_framework_path(p):
    return bool(FRAMEWORK_FILE_RE.search(str(p)))


def parse_exclude_folders(
    user_exclude_folders=None, use_defaults=True, include_folders=None
):
    exclude_folders = set()

    if use_defaults:
        exclude_folders.update(DEFAULT_EXCLUDE_FOLDERS)

    if user_exclude_folders:
        exclude_folders.update(user_exclude_folders)

    if include_folders:
        for folder in include_folders:
            exclude_folders.discard(folder)

    return exclude_folders



import ast


class TaintVisitor(ast.NodeVisitor):
    def __init__(self, file_path, findings):
        self.file_path = file_path
        self.findings = findings
        self.env_stack = [{}]
        self.sources = {"input", "request"}
        self.request_obj = "request"

    def _push(self):
        self.env_stack.append({})

    def _pop(self):
        if self.env_stack:
            self.env_stack.pop()

    def _set(self, name, tainted):
        if not self.env_stack:
            self._push()
        self.env_stack[-1][name] = bool(tainted)

    def _get(self, name):
        for env in reversed(self.env_stack):
            if name in env:
                return env[name]
        return False

    def is_tainted(self, node):
        if node is None:
            return False

        if isinstance(node, ast.JoinedStr):
            return any(
                isinstance(v, ast.FormattedValue) and self.is_tainted(v.value)
                for v in node.values
            )
        if isinstance(node, ast.BinOp) and isinstance(node.op, (ast.Add, ast.Mod)):
            return self.is_tainted(node.left) or self.is_tainted(node.right)
        if (
            isinstance(node, ast.Call)
            and isinstance(node.func, ast.Attribute)
            and node.func.attr == "format"
        ):
            return True

        if (
            isinstance(node, ast.Call)
            and isinstance(node.func, ast.Name)
            and node.func.id in self.sources
        ):
            return True

        if isinstance(node, (ast.Attribute, ast.Subscript)):
            base = node.value
            while isinstance(base, (ast.Attribute, ast.Subscript)):
                base = base.value
            if isinstance(base, ast.Name) and base.id == self.request_obj:
                return True
            return self.is_tainted(node.value)

        if isinstance(node, ast.Name):
            return self._get(node.id)

        if isinstance(node, ast.Call):
            if isinstance(node.func, ast.Attribute):
                if self.is_tainted(node.func.value):
                    return True

            if any(self.is_tainted(arg) for arg in node.args):
                return True

            if any(self.is_tainted(k.value) for k in node.keywords):
                return True

        return False

    def generic_visit(self, node):
        for field, value in ast.iter_fields(node):
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, ast.AST):
                        self.visit(item)
            elif isinstance(value, ast.AST):
                self.visit(value)

    def visit_FunctionDef(self, node):
        self._push()
        self.generic_visit(node)
        self._pop()

    def visit_AsyncFunctionDef(self, node):
        self._push()
        self.generic_visit(node)
        self._pop()

    def visit_Assign(self, node):
        t = self.is_tainted(node.value)
        for tgt in node.targets:
            if isinstance(tgt, ast.Name):
                self._set(tgt.id, t)
        self.generic_visit(node)

    def visit_AnnAssign(self, node):
        if node.value:
            t = self.is_tainted(node.value)
            if isinstance(node.target, ast.Name):
                self._set(node.target.id, t)
        self.generic_visit(node)
